# --------------------------------------------------------
# Neighborhood Attention Transformer
# Licensed under The MIT License
# Written by Ali Hassani
# --------------------------------------------------------

# Modified by Jitesh Jain

import torch
import torch.nn as nn
from custom_timm.models.layers import DropPath
from custom_detectron2.modeling import BACKBONE_REGISTRY, Backbone, ShapeSpec

class NeighborhoodAttention(nn.Module):
    """
    Neighborhood Attention 2D Module
    """

    def __init__(
        self,
        dim,
        num_heads,
        kernel_size,
        dilation=1,
        bias=True,
        qkv_bias=True,
        qk_scale=None,
        attn_drop=0.0,
        proj_drop=0.0,
    ):
        super().__init__()


    def forward(self, x):

        return x

    def extra_repr(self) -> str:
        return (
            f"head_dim={self.head_dim}, num_heads={self.num_heads}, "
            + f"kernel_size={self.kernel_size}, dilation={self.dilation}, "
            + f"rel_pos_bias={self.rpb is not None}"
        )

class ConvTokenizer(nn.Module):
    def __init__(self, in_chans=3, embed_dim=96, norm_layer=None):
        super().__init__()
        self.proj = nn.Sequential(
            nn.Conv2d(in_chans, embed_dim // 2, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
            nn.Conv2d(embed_dim // 2, embed_dim, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
        )
        if norm_layer is not None:
            self.norm = norm_layer(embed_dim)
        else:
            self.norm = None

    def forward(self, x):
        x = self.proj(x).permute(0, 2, 3, 1)
        if self.norm is not None:
            x = self.norm(x)
        return x


class ConvDownsampler(nn.Module):
    def __init__(self, dim, norm_layer=nn.LayerNorm):
        super().__init__()
        self.reduction = nn.Conv2d(dim, 2 * dim, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), bias=False)
        self.norm = norm_layer(2 * dim)

    def forward(self, x):
        x = self.reduction(x.permute(0, 3, 1, 2)).permute(0, 2, 3, 1)
        x = self.norm(x)
        return x


class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class NATLayer(nn.Module):
    def __init__(self, dim, num_heads, kernel_size=7, dilation=None,
                 mlp_ratio=4., qkv_bias=True, qk_scale=None, drop=0., attn_drop=0., drop_path=0.,
                 act_layer=nn.GELU, norm_layer=nn.LayerNorm, layer_scale=None):
        super().__init__()
        self.dim = dim
        self.num_heads = num_heads
        self.mlp_ratio = mlp_ratio

        self.norm1 = norm_layer(dim)
        self.attn = NeighborhoodAttention(
            dim, kernel_size=kernel_size, dilation=dilation, num_heads=num_heads,
            qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)

        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        self.mlp = Mlp(in_features=dim, hidden_features=int(dim * mlp_ratio), act_layer=act_layer, drop=drop)
        self.layer_scale = False
        if layer_scale is not None and type(layer_scale) in [int, float]:
            self.layer_scale = True
            self.gamma1 = nn.Parameter(layer_scale * torch.ones(dim), requires_grad=True)
            self.gamma2 = nn.Parameter(layer_scale * torch.ones(dim), requires_grad=True)

    def forward(self, x):
        if not self.layer_scale:
            shortcut = x
            x = self.norm1(x)
            x = self.attn(x)
            x = shortcut + self.drop_path(x)
            x = x + self.drop_path(self.mlp(self.norm2(x)))
            return x
        shortcut = x
        x = self.norm1(x)
        x = self.attn(x)
        x = shortcut + self.drop_path(self.gamma1 * x)
        x = x + self.drop_path(self.gamma2 * self.mlp(self.norm2(x)))
        return x



class NATBlock(nn.Module):
    def __init__(self, dim, depth, num_heads, kernel_size,  dilations=None,
                 downsample=True,
                 mlp_ratio=4., qkv_bias=True, qk_scale=None, drop=0., attn_drop=0.,
                 drop_path=0., norm_layer=nn.LayerNorm, layer_scale=None):
        super().__init__()
        self.dim = dim
        self.depth = depth

        self.blocks = nn.ModuleList([
            NATLayer(dim=dim,
                     num_heads=num_heads,
                     kernel_size=kernel_size,
                     dilation=None if dilations is None else dilations[i],
                     mlp_ratio=mlp_ratio,
                     qkv_bias=qkv_bias, qk_scale=qk_scale,
                     drop=drop, attn_drop=attn_drop,
                     drop_path=drop_path[i] if isinstance(drop_path, list) else drop_path,
                     norm_layer=norm_layer,
                     layer_scale=layer_scale)
            for i in range(depth)])

        self.downsample = None if not downsample else ConvDownsampler(dim=dim, norm_layer=norm_layer)

    def forward(self, x):
        for blk in self.blocks:
            x = blk(x)
        if self.downsample is None:
            return x, x
        return self.downsample(x), x


class DiNAT(nn.Module):
    def __init__(self,
                 embed_dim,
                 mlp_ratio,
                 depths,
                 num_heads,
                 drop_path_rate=0.2,
                 in_chans=3,
                 kernel_size=7,
                 dilations=None,
                 out_indices=(0, 1, 2, 3),
                 qkv_bias=True,
                 qk_scale=None,
                 drop_rate=0.,
                 attn_drop_rate=0.,
                 norm_layer=nn.LayerNorm,
                 frozen_stages=-1,
                 layer_scale=None,
                 **kwargs):
        super().__init__()
        self.num_levels = len(depths)
        self.embed_dim = embed_dim
        self.num_features = [int(embed_dim * 2 ** i) for i in range(self.num_levels)]
        self.mlp_ratio = mlp_ratio

        self.patch_embed = ConvTokenizer(in_chans=in_chans, embed_dim=embed_dim, norm_layer=norm_layer)

        self.pos_drop = nn.Dropout(p=drop_rate)

        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]
        self.levels = nn.ModuleList()
        for i in range(self.num_levels):
            level = NATBlock(dim=int(embed_dim * 2 ** i),
                             depth=depths[i],
                             num_heads=num_heads[i],
                             kernel_size=kernel_size,
                             dilations=None if dilations is None else dilations[i],
                             mlp_ratio=self.mlp_ratio,
                             qkv_bias=qkv_bias, qk_scale=qk_scale,
                             drop=drop_rate, attn_drop=attn_drop_rate,
                             drop_path=dpr[sum(depths[:i]):sum(depths[:i + 1])],
                             norm_layer=norm_layer,
                             downsample=(i < self.num_levels - 1),
                             layer_scale=layer_scale)
            self.levels.append(level)

        # add a norm layer for each output
        self.out_indices = out_indices
        for i_layer in self.out_indices:
            layer = norm_layer(self.num_features[i_layer])
            layer_name = f'norm{i_layer}'
            self.add_module(layer_name, layer)

        self.frozen_stages = frozen_stages

    def _freeze_stages(self):
        if self.frozen_stages >= 0:
            self.patch_embed.eval()
            for param in self.patch_embed.parameters():
                param.requires_grad = False

        if self.frozen_stages >= 2:
            for i in range(0, self.frozen_stages - 1):
                m = self.network[i]
                m.eval()
                for param in m.parameters():
                    param.requires_grad = False

    def train(self, mode=True):
        super(DiNAT, self).train(mode)
        self._freeze_stages()

    def forward_embeddings(self, x):
        x = self.patch_embed(x)
        return x

    def forward_tokens(self, x):
        outs = {}
        for idx, level in enumerate(self.levels):
            x, xo = level(x)
            if idx in self.out_indices:
                norm_layer = getattr(self, f'norm{idx}')
                x_out = norm_layer(xo)
                outs["res{}".format(idx + 2)] = x_out.permute(0, 3, 1, 2).contiguous()
        return outs

    def forward(self, x):
        x = self.forward_embeddings(x)
        return self.forward_tokens(x)


@BACKBONE_REGISTRY.register()
class D2DiNAT(DiNAT, Backbone):
    def __init__(self, cfg, input_shape):
        
        embed_dim = cfg.MODEL.DiNAT.EMBED_DIM
        mlp_ratio = cfg.MODEL.DiNAT.MLP_RATIO
        depths = cfg.MODEL.DiNAT.DEPTHS
        num_heads = cfg.MODEL.DiNAT.NUM_HEADS
        drop_path_rate = cfg.MODEL.DiNAT.DROP_PATH_RATE
        kernel_size = cfg.MODEL.DiNAT.KERNEL_SIZE
        out_indices = cfg.MODEL.DiNAT.OUT_INDICES
        dilations = cfg.MODEL.DiNAT.DILATIONS

        super().__init__(
            embed_dim=embed_dim,
            mlp_ratio=mlp_ratio,
            depths=depths,
            num_heads=num_heads,
            drop_path_rate=drop_path_rate,
            kernel_size=kernel_size,
            out_indices=out_indices,
            dilations=dilations,
        )

        self._out_features = cfg.MODEL.DiNAT.OUT_FEATURES

        self._out_feature_strides = {
            "res2": 4,
            "res3": 8,
            "res4": 16,
            "res5": 32,
        }
        self._out_feature_channels = {
            "res2": self.num_features[0],
            "res3": self.num_features[1],
            "res4": self.num_features[2],
            "res5": self.num_features[3],
        }

    def forward(self, x):
        """
        Args:
            x: Tensor of shape (N,C,H,W). H, W must be a multiple of ``self.size_divisibility``.
        Returns:
            dict[str->Tensor]: names and the corresponding features
        """
        assert (
            x.dim() == 4
        ), f"DiNAT takes an input of shape (N, C, H, W). Got {x.shape} instead!"
        outputs = {}
        y = super().forward(x)
        for k in y.keys():
            if k in self._out_features:
                outputs[k] = y[k]
        return outputs

    def output_shape(self):
        return {
            name: ShapeSpec(
                channels=self._out_feature_channels[name], stride=self._out_feature_strides[name]
            )
            for name in self._out_features
        }

    @property
    def size_divisibility(self):
        return 32



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright (c) @comfyanonymous
# Project Repository: https://github.com/comfyanonymous/ComfyUI

import os
import time
from typing import Optional, Union
import os

from nodetool.common.environment import Environment

supported_pt_extensions = set([".ckpt", ".pt", ".bin", ".pth", ".safetensors"])
filename_list_cache: dict[str, tuple[list[str], dict[str, float], float]] = {}
data_dir = Environment.get_comfy_folder()
models_dir = os.path.join(data_dir, "models")

custom_nodes_directory = os.path.join(
    os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
    "comfy_custom_nodes",
)


def model_path(name: str) -> str:
    return os.path.join(models_dir, name)


folder_names_and_paths: dict[str, tuple[list[str], set[str]]] = {
    "annotator": ([model_path("annotator")], supported_pt_extensions),
    "checkpoints": ([model_path("checkpoints")], supported_pt_extensions),
    "configs": ([model_path("configs")], set([".yaml"])),
    "loras": ([model_path("loras")], supported_pt_extensions),
    "vae": ([model_path("vae")], supported_pt_extensions),
    "clip": ([model_path("clip")], supported_pt_extensions),
    "unet": ([model_path("unet")], supported_pt_extensions),
    "clip_vision": ([model_path("clip_vision")], supported_pt_extensions),
    "style_models": ([model_path("style_models")], supported_pt_extensions),
    "embeddings": ([model_path("embeddings")], supported_pt_extensions),
    "vae_approx": ([model_path("vae_approx")], supported_pt_extensions),
    "controlnet": (
        [model_path("controlnet"), model_path("t2i_adapter")],
        supported_pt_extensions,
    ),
    "gligen": ([model_path("gligen")], supported_pt_extensions),
    "upscale_models": ([model_path("upscale_models")], supported_pt_extensions),
    "custom_nodes": ([custom_nodes_directory], set()),
    "hypernetworks": ([model_path("hypernetworks")], supported_pt_extensions),
    "classifiers": ([model_path("classifiers")], {""}),
}

print("Initializing comfy folder paths")
for name in ["checkpoints", "loras"]:
    print("Searching for ", name)
    for file in os.listdir(folder_names_and_paths[name][0][0]):
        print("Found ", file)


def add_model_folder_path(folder_name: str, full_folder_path: str):
    """
    Add a new folder path for a specific model type.

    Args:
        folder_name (str): The name of the model type.
        full_folder_path (str): The full path to the folder containing the models.
    """
    global folder_names_and_paths
    if folder_name in folder_names_and_paths:
        folder_names_and_paths[folder_name][0].append(full_folder_path)
    else:
        folder_names_and_paths[folder_name] = ([full_folder_path], set())


def get_folder_paths(folder_name: str) -> list[str]:
    """
    Get the folder paths for a specific model type.

    Args:
        folder_name (str): The name of the model type.

    Returns:
        list[str]: The list of folder paths for the specified model type.
    """
    return folder_names_and_paths[folder_name][0][:]


def recursive_search(
    directory: str, excluded_dir_names: Optional[list[str]] = None
) -> tuple[list[str], dict[str, float]]:
    """
    Recursively search for files in a directory and its subdirectories.

    Args:
        directory (str): The directory to search.
        excluded_dir_names (Optional[list[str]], optional): A list of directory names to exclude from the search. Defaults to None.

    Returns:
        tuple[list[str], dict[str, float]]: A tuple containing the list of file paths and a dictionary of directory paths and their modification times.
    """
    if not os.path.isdir(directory):
        return [], {}

    if excluded_dir_names is None:
        excluded_dir_names = []

    result: list[str] = []
    dirs: dict[str, float] = {directory: os.path.getmtime(directory)}
    for dirpath, subdirs, filenames in os.walk(
        directory, followlinks=True, topdown=True
    ):
        subdirs[:] = [d for d in subdirs if d not in excluded_dir_names]
        for file_name in filenames:
            relative_path = os.path.relpath(os.path.join(dirpath, file_name), directory)
            result.append(relative_path)
        for d in subdirs:
            path = os.path.join(dirpath, d)
            dirs[path] = os.path.getmtime(path)
    return result, dirs


def filter_files_extensions(files: list[str], extensions: list[str]) -> list[str]:
    """
    Filter a list of file paths based on their extensions.

    Args:
        files (list[str]): The list of file paths.
        extensions (list[str]): The list of allowed extensions.

    Returns:
        list[str]: The filtered list of file paths.
    """
    return sorted(
        list(
            filter(
                lambda a: os.path.splitext(a)[-1].lower() in extensions
                or len(extensions) == 0,
                files,
            )
        )
    )


def get_full_path(folder_name: str, filename: str) -> str | None:
    """
    Get the full path of a file given its folder name and filename.

    Args:
        folder_name (str): The name of the folder.
        filename (str): The name of the file.

    Returns:
        str | None: The full path of the file if found, None otherwise.
    """
    global folder_names_and_paths
    if folder_name not in folder_names_and_paths:
        return None
    folders = folder_names_and_paths[folder_name]
    filename = os.path.relpath(os.path.join("/", filename), "/")
    for x in folders[0]:
        full_path = os.path.join(x, filename)
        if os.path.isfile(full_path):
            return full_path

    return None


def get_filename_list_(folder_name: str) -> tuple[list[str], dict[str, float], float]:
    """
    Get the list of filenames and folder modification times for a specific folder.

    Args:
        folder_name (str): The name of the folder.

    Returns:
        tuple[list[str], dict[str, float], float]: A tuple containing the list of filenames, a dictionary of folder paths and their modification times, and the current time.
    """
    global folder_names_and_paths
    output_list: set[str] = set()
    folders = folder_names_and_paths[folder_name]
    output_folders: dict[str, float] = {}

    for x in folders[0]:
        files, folders_all = recursive_search(x, excluded_dir_names=[".git"])
        output_list.update(filter_files_extensions(files, list(folders[1])))
        output_folders = {**output_folders, **folders_all}

    return (sorted(list(output_list)), output_folders, time.perf_counter())


def cached_filename_list_(
    folder_name: str,
) -> Optional[tuple[list[str], dict[str, float], float]]:
    """
    Get the cached list of filenames and folder modification times for a
    Args:
        folder_name (str): The name of the folder.

    Returns:
        Optional[Tuple[List[str], Dict[str, float], float]]: A tuple containing the cached list of filenames, a dictionary of folder paths and their modification times, and the cache timestamp if available, None otherwise.
    """
    global filename_list_cache
    global folder_names_and_paths
    if folder_name not in filename_list_cache:
        return None
    out = filename_list_cache[folder_name]
    if time.perf_counter() < (out[2] + 0.5):
        return out
    for x in out[1]:
        time_modified = out[1][x]
        folder = x
        if os.path.getmtime(folder) != time_modified:
            return None

    folders = folder_names_and_paths[folder_name]
    for x in folders[0]:
        if os.path.isdir(x):
            if x not in out[1]:
                return None

    return out


def get_filename_list(folder_name):
    """
    Get the list of filenames for a specific folder, using the cache if available.

    Args:
        folder_name (str): The name of the folder.

    Returns:
        list[str]: The list of filenames for the specified folder.
    """
    out = cached_filename_list_(folder_name)
    if out is None:
        out = get_filename_list_(folder_name)
        global filename_list_cache
        filename_list_cache[folder_name] = out
    return list(out[0])



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class ExtractVideoFrames(GraphNode):
    video: VideoRef | GraphNode | tuple[GraphNode, str] = Field(default=PydanticUndefined, description='The input video to adjust the brightness for.')
    start: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The frame to start extracting from.')
    end: int | GraphNode | tuple[GraphNode, str] = Field(default=-1, description='The frame to stop extracting from.')
    @classmethod
    def get_node_type(cls): return "nodetool.video.ExtractVideoFrames"



class FramesToVideo(GraphNode):
    frames: list[ImageRef] | GraphNode | tuple[GraphNode, str] = Field(default=[], description=None)
    fps: float | GraphNode | tuple[GraphNode, str] = Field(default=30, description='The FPS of the output video.')
    @classmethod
    def get_node_type(cls): return "nodetool.video.FramesToVideo"



class SaveVideo(GraphNode):
    value: VideoRef | GraphNode | tuple[GraphNode, str] = Field(default=VideoRef(type='video', uri='', asset_id=None, temp_id=None, duration=None, format=None), description=None)
    folder: FolderRef | GraphNode | tuple[GraphNode, str] = Field(default=FolderRef(type='folder', uri='', asset_id=None, temp_id=None), description='Name of the output folder.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='video', description='Name of the output video.')
    @classmethod
    def get_node_type(cls): return "nodetool.video.SaveVideo"



class VideoFps(GraphNode):
    video: VideoRef | GraphNode | tuple[GraphNode, str] = Field(default=PydanticUndefined, description='The input video to adjust the brightness for.')
    @classmethod
    def get_node_type(cls): return "nodetool.video.VideoFps"





"""
Copyright (c) Microsoft Corporation.
Licensed under the MIT license.

"""

import torch
import custom_mesh_graphormer.modeling.data.config as cfg
from comfy.model_management import get_torch_device
device = get_torch_device()

class Graphormer_Hand_Network(torch.nn.Module):
    '''
    End-to-end Graphormer network for hand pose and mesh reconstruction from a single image.
    '''
    def __init__(self, args, config, backbone, trans_encoder):
        super(Graphormer_Hand_Network, self).__init__()
        self.config = config
        self.backbone = backbone
        self.trans_encoder = trans_encoder
        self.upsampling = torch.nn.Linear(195, 778)
        self.cam_param_fc = torch.nn.Linear(3, 1)
        self.cam_param_fc2 = torch.nn.Linear(195+21, 150) 
        self.cam_param_fc3 = torch.nn.Linear(150, 3)
        self.grid_feat_dim = torch.nn.Linear(1024, 2051)

    def forward(self, images, mesh_model, mesh_sampler, meta_masks=None, is_train=False):
        batch_size = images.size(0)
        # Generate T-pose template mesh
        template_pose = torch.zeros((1,48))
        template_pose = template_pose.to(device)
        template_betas = torch.zeros((1,10)).to(device)
        template_vertices, template_3d_joints = mesh_model.layer(template_pose, template_betas)
        template_vertices = template_vertices/1000.0
        template_3d_joints = template_3d_joints/1000.0

        template_vertices_sub = mesh_sampler.downsample(template_vertices)

        # normalize
        template_root = template_3d_joints[:,cfg.J_NAME.index('Wrist'),:]
        template_3d_joints = template_3d_joints - template_root[:, None, :]
        template_vertices = template_vertices - template_root[:, None, :]
        template_vertices_sub = template_vertices_sub - template_root[:, None, :]
        num_joints = template_3d_joints.shape[1]

        # concatinate template joints and template vertices, and then duplicate to batch size
        ref_vertices = torch.cat([template_3d_joints, template_vertices_sub],dim=1)
        ref_vertices = ref_vertices.expand(batch_size, -1, -1)

        # extract grid features and global image features using a CNN backbone
        image_feat, grid_feat = self.backbone(images)
        # concatinate image feat and mesh template
        image_feat = image_feat.view(batch_size, 1, 2048).expand(-1, ref_vertices.shape[-2], -1)
        # process grid features
        grid_feat = torch.flatten(grid_feat, start_dim=2)
        grid_feat = grid_feat.transpose(1,2)
        grid_feat = self.grid_feat_dim(grid_feat)
        # concatinate image feat and template mesh to form the joint/vertex queries
        features = torch.cat([ref_vertices, image_feat], dim=2)
        # prepare input tokens including joint/vertex queries and grid features
        features = torch.cat([features, grid_feat],dim=1)

        if is_train==True:
            # apply mask vertex/joint modeling
            # meta_masks is a tensor of all the masks, randomly generated in dataloader
            # we pre-define a [MASK] token, which is a floating-value vector with 0.01s  
            special_token = torch.ones_like(features[:,:-49,:]).to(device)*0.01
            features[:,:-49,:] = features[:,:-49,:]*meta_masks + special_token*(1-meta_masks)

        # forward pass
        if self.config.output_attentions==True:
            features, hidden_states, att = self.trans_encoder(features)
        else:
            features = self.trans_encoder(features)

        pred_3d_joints = features[:,:num_joints,:]
        pred_vertices_sub = features[:,num_joints:-49,:]

        # learn camera parameters
        x = self.cam_param_fc(features[:,:-49,:])
        x = x.transpose(1,2)
        x = self.cam_param_fc2(x)
        x = self.cam_param_fc3(x)
        cam_param = x.transpose(1,2)
        cam_param = cam_param.squeeze()

        temp_transpose = pred_vertices_sub.transpose(1,2)
        pred_vertices = self.upsampling(temp_transpose)
        pred_vertices = pred_vertices.transpose(1,2)

        if self.config.output_attentions==True:
            return cam_param, pred_3d_joints, pred_vertices_sub, pred_vertices, hidden_states, att
        else:
            return cam_param, pred_3d_joints, pred_vertices_sub, pred_vertices


from datetime import datetime
from typing import Literal

from pydantic import BaseModel
from nodetool.models.base_model import DBModel, DBField, create_time_ordered_uuid


class MessageTextContent(BaseModel):
    type: Literal["text"] = "text"
    text: str = ""


class ImageUrl(BaseModel):
    url: str = ""


class MessageImageContent(BaseModel):
    type: Literal["image_url"] = "image_url"
    image_url: ImageUrl = ImageUrl(url="")


MessageContent = MessageTextContent | MessageImageContent


class Message(DBModel):
    @classmethod
    def get_table_schema(cls):
        return {
            "table_name": "nodetool_messages",
            "key_schema": {"id": "HASH"},
            "attribute_definitions": {"id": "S", "thread_id": "S"},
            "global_secondary_indexes": {
                "nodetool_message_thread_index": {
                    "thread_id": "HASH",
                },
            },
        }

    id: str = DBField(hash_key=True)
    thread_id: str = DBField(default="")
    user_id: str = DBField(default="")
    tool_call_id: str | None = DBField(default=None)
    role: str = DBField(default="")
    name: str = DBField(default="")
    content: str | list[MessageContent] | None = DBField(default=None)
    tool_calls: list[dict] | None = DBField(default=None)
    created_at: datetime = DBField(default_factory=datetime.now)

    @classmethod
    def create(cls, thread_id: str, user_id: str, **kwargs) -> "Message":
        return super().create(
            id=create_time_ordered_uuid(),
            thread_id=thread_id,
            user_id=user_id,
            **kwargs,
        )

    @classmethod
    def paginate(
        cls,
        thread_id: str | None = None,
        limit: int = 10,
        start_key: str | None = None,
        reverse: bool = False,
    ):
        return cls.query(
            condition="thread_id = :thread_id",
            values={":thread_id": thread_id},
            index="nodetool_message_thread_index",
            limit=limit,
            start_key=start_key,
            reverse=reverse,
        )



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class AudioFolder(GraphNode):
    label: str | GraphNode | tuple[GraphNode, str] = Field(default='Input Label', description='The label for this input node.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The parameter name for the workflow.')
    description: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The description for this input node.')
    folder: FolderRef | GraphNode | tuple[GraphNode, str] = Field(default=FolderRef(type='folder', uri='', asset_id=None, temp_id=None), description='The folder to use as input.')
    limit: int | GraphNode | tuple[GraphNode, str] = Field(default=1000, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.input.AudioFolder"



class AudioInput(GraphNode):
    label: str | GraphNode | tuple[GraphNode, str] = Field(default='Input Label', description='The label for this input node.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The parameter name for the workflow.')
    description: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The description for this input node.')
    value: AudioRef | GraphNode | tuple[GraphNode, str] = Field(default=AudioRef(type='audio', uri='', asset_id=None, temp_id=None), description='The audio to use as input.')
    @classmethod
    def get_node_type(cls): return "nodetool.input.AudioInput"



class BooleanInput(GraphNode):
    label: str | GraphNode | tuple[GraphNode, str] = Field(default='Input Label', description='The label for this input node.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The parameter name for the workflow.')
    description: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The description for this input node.')
    value: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.input.BooleanInput"



class ChatInput(GraphNode):
    label: str | GraphNode | tuple[GraphNode, str] = Field(default='Input Label', description='The label for this input node.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The parameter name for the workflow.')
    description: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The description for this input node.')
    value: str | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.input.ChatInput"



class ComfyImageInput(GraphNode):
    label: str | GraphNode | tuple[GraphNode, str] = Field(default='Input Label', description='The label for this input node.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The parameter name for the workflow.')
    description: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The description for this input node.')
    value: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to use as input.')
    @classmethod
    def get_node_type(cls): return "nodetool.input.ComfyImageInput"



class FloatInput(GraphNode):
    label: str | GraphNode | tuple[GraphNode, str] = Field(default='Input Label', description='The label for this input node.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The parameter name for the workflow.')
    description: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The description for this input node.')
    value: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    min: float | GraphNode | tuple[GraphNode, str] = Field(default=0, description=None)
    max: float | GraphNode | tuple[GraphNode, str] = Field(default=100, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.input.FloatInput"



class Folder(GraphNode):
    label: str | GraphNode | tuple[GraphNode, str] = Field(default='Input Label', description='The label for this input node.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The parameter name for the workflow.')
    description: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The description for this input node.')
    folder: FolderRef | GraphNode | tuple[GraphNode, str] = Field(default=FolderRef(type='folder', uri='', asset_id=None, temp_id=None), description='The folder to use as input.')
    limit: int | GraphNode | tuple[GraphNode, str] = Field(default=1000, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.input.Folder"



class GroupInput(GraphNode):
    items: list[Any] | GraphNode | tuple[GraphNode, str] = Field(default=[], description=None)
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.input.GroupInput"



class ImageFolder(GraphNode):
    label: str | GraphNode | tuple[GraphNode, str] = Field(default='Input Label', description='The label for this input node.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The parameter name for the workflow.')
    description: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The description for this input node.')
    folder: FolderRef | GraphNode | tuple[GraphNode, str] = Field(default=FolderRef(type='folder', uri='', asset_id=None, temp_id=None), description='The folder to use as input.')
    limit: int | GraphNode | tuple[GraphNode, str] = Field(default=1000, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.input.ImageFolder"



class ImageInput(GraphNode):
    label: str | GraphNode | tuple[GraphNode, str] = Field(default='Input Label', description='The label for this input node.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The parameter name for the workflow.')
    description: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The description for this input node.')
    value: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to use as input.')
    @classmethod
    def get_node_type(cls): return "nodetool.input.ImageInput"



class IntegerInput(GraphNode):
    label: str | GraphNode | tuple[GraphNode, str] = Field(default='Input Label', description='The label for this input node.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The parameter name for the workflow.')
    description: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The description for this input node.')
    value: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description=None)
    min: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description=None)
    max: int | GraphNode | tuple[GraphNode, str] = Field(default=100, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.input.IntegerInput"



class StringInput(GraphNode):
    label: str | GraphNode | tuple[GraphNode, str] = Field(default='Input Label', description='The label for this input node.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The parameter name for the workflow.')
    description: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The description for this input node.')
    value: str | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.input.StringInput"



class TextFolder(GraphNode):
    label: str | GraphNode | tuple[GraphNode, str] = Field(default='Input Label', description='The label for this input node.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The parameter name for the workflow.')
    description: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The description for this input node.')
    folder: FolderRef | GraphNode | tuple[GraphNode, str] = Field(default=FolderRef(type='folder', uri='', asset_id=None, temp_id=None), description='The folder to use as input.')
    limit: int | GraphNode | tuple[GraphNode, str] = Field(default=1000, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.input.TextFolder"



class TextInput(GraphNode):
    label: str | GraphNode | tuple[GraphNode, str] = Field(default='Input Label', description='The label for this input node.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The parameter name for the workflow.')
    description: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The description for this input node.')
    value: TextRef | GraphNode | tuple[GraphNode, str] = Field(default=TextRef(type='text', uri='', asset_id=None, temp_id=None), description='The text to use as input.')
    @classmethod
    def get_node_type(cls): return "nodetool.input.TextInput"



class VideoFolder(GraphNode):
    label: str | GraphNode | tuple[GraphNode, str] = Field(default='Input Label', description='The label for this input node.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The parameter name for the workflow.')
    description: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The description for this input node.')
    folder: FolderRef | GraphNode | tuple[GraphNode, str] = Field(default=FolderRef(type='folder', uri='', asset_id=None, temp_id=None), description='The folder to use as input.')
    limit: int | GraphNode | tuple[GraphNode, str] = Field(default=1000, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.input.VideoFolder"



class VideoInput(GraphNode):
    label: str | GraphNode | tuple[GraphNode, str] = Field(default='Input Label', description='The label for this input node.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The parameter name for the workflow.')
    description: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The description for this input node.')
    value: VideoRef | GraphNode | tuple[GraphNode, str] = Field(default=VideoRef(type='video', uri='', asset_id=None, temp_id=None, duration=None, format=None), description='The video to use as input.')
    @classmethod
    def get_node_type(cls): return "nodetool.input.VideoInput"





from typing import Tuple
import math
import numpy as np
from enum import IntEnum
from typing import List, Tuple, Union
import torch
from torch.nn import functional as F
import logging
import cv2

Image = np.ndarray
Boxes = torch.Tensor
ImageSizeType = Tuple[int, int]
_RawBoxType = Union[List[float], Tuple[float, ...], torch.Tensor, np.ndarray]
IntTupleBox = Tuple[int, int, int, int]

class BoxMode(IntEnum):
    """
    Enum of different ways to represent a box.
    """

    XYXY_ABS = 0
    """
    (x0, y0, x1, y1) in absolute floating points coordinates.
    The coordinates in range [0, width or height].
    """
    XYWH_ABS = 1
    """
    (x0, y0, w, h) in absolute floating points coordinates.
    """
    XYXY_REL = 2
    """
    Not yet supported!
    (x0, y0, x1, y1) in range [0, 1]. They are relative to the size of the image.
    """
    XYWH_REL = 3
    """
    Not yet supported!
    (x0, y0, w, h) in range [0, 1]. They are relative to the size of the image.
    """
    XYWHA_ABS = 4
    """
    (xc, yc, w, h, a) in absolute floating points coordinates.
    (xc, yc) is the center of the rotated box, and the angle a is in degrees ccw.
    """

    @staticmethod
    def convert(box: _RawBoxType, from_mode: "BoxMode", to_mode: "BoxMode") -> _RawBoxType:
        """
        Args:
            box: can be a k-tuple, k-list or an Nxk array/tensor, where k = 4 or 5
            from_mode, to_mode (BoxMode)

        Returns:
            The converted box of the same type.
        """
        if from_mode == to_mode:
            return box

        original_type = type(box)
        is_numpy = isinstance(box, np.ndarray)
        single_box = isinstance(box, (list, tuple))
        if single_box:
            assert len(box) == 4 or len(box) == 5, (
                "BoxMode.convert takes either a k-tuple/list or an Nxk array/tensor,"
                " where k == 4 or 5"
            )
            arr = torch.tensor(box)[None, :]
        else:
            # avoid modifying the input box
            if is_numpy:
                arr = torch.from_numpy(np.asarray(box)).clone()
            else:
                arr = box.clone()

        assert to_mode not in [BoxMode.XYXY_REL, BoxMode.XYWH_REL] and from_mode not in [
            BoxMode.XYXY_REL,
            BoxMode.XYWH_REL,
        ], "Relative mode not yet supported!"

        if from_mode == BoxMode.XYWHA_ABS and to_mode == BoxMode.XYXY_ABS:
            assert (
                arr.shape[-1] == 5
            ), "The last dimension of input shape must be 5 for XYWHA format"
            original_dtype = arr.dtype
            arr = arr.double()

            w = arr[:, 2]
            h = arr[:, 3]
            a = arr[:, 4]
            c = torch.abs(torch.cos(a * math.pi / 180.0))
            s = torch.abs(torch.sin(a * math.pi / 180.0))
            # This basically computes the horizontal bounding rectangle of the rotated box
            new_w = c * w + s * h
            new_h = c * h + s * w

            # convert center to top-left corner
            arr[:, 0] -= new_w / 2.0
            arr[:, 1] -= new_h / 2.0
            # bottom-right corner
            arr[:, 2] = arr[:, 0] + new_w
            arr[:, 3] = arr[:, 1] + new_h

            arr = arr[:, :4].to(dtype=original_dtype)
        elif from_mode == BoxMode.XYWH_ABS and to_mode == BoxMode.XYWHA_ABS:
            original_dtype = arr.dtype
            arr = arr.double()
            arr[:, 0] += arr[:, 2] / 2.0
            arr[:, 1] += arr[:, 3] / 2.0
            angles = torch.zeros((arr.shape[0], 1), dtype=arr.dtype)
            arr = torch.cat((arr, angles), axis=1).to(dtype=original_dtype)
        else:
            if to_mode == BoxMode.XYXY_ABS and from_mode == BoxMode.XYWH_ABS:
                arr[:, 2] += arr[:, 0]
                arr[:, 3] += arr[:, 1]
            elif from_mode == BoxMode.XYXY_ABS and to_mode == BoxMode.XYWH_ABS:
                arr[:, 2] -= arr[:, 0]
                arr[:, 3] -= arr[:, 1]
            else:
                raise NotImplementedError(
                    "Conversion from BoxMode {} to {} is not supported yet".format(
                        from_mode, to_mode
                    )
                )

        if single_box:
            return original_type(arr.flatten().tolist())
        if is_numpy:
            return arr.numpy()
        else:
            return arr

class MatrixVisualizer:
    """
    Base visualizer for matrix data
    """

    def __init__(
        self,
        inplace=True,
        cmap=cv2.COLORMAP_PARULA,
        val_scale=1.0,
        alpha=0.7,
        interp_method_matrix=cv2.INTER_LINEAR,
        interp_method_mask=cv2.INTER_NEAREST,
    ):
        self.inplace = inplace
        self.cmap = cmap
        self.val_scale = val_scale
        self.alpha = alpha
        self.interp_method_matrix = interp_method_matrix
        self.interp_method_mask = interp_method_mask

    def visualize(self, image_bgr, mask, matrix, bbox_xywh):
        self._check_image(image_bgr)
        self._check_mask_matrix(mask, matrix)
        if self.inplace:
            image_target_bgr = image_bgr
        else:
            image_target_bgr = image_bgr * 0
        x, y, w, h = [int(v) for v in bbox_xywh]
        if w <= 0 or h <= 0:
            return image_bgr
        mask, matrix = self._resize(mask, matrix, w, h)
        mask_bg = np.tile((mask == 0)[:, :, np.newaxis], [1, 1, 3])
        matrix_scaled = matrix.astype(np.float32) * self.val_scale
        _EPSILON = 1e-6
        if np.any(matrix_scaled > 255 + _EPSILON):
            logger = logging.getLogger(__name__)
            logger.warning(
                f"Matrix has values > {255 + _EPSILON} after " f"scaling, clipping to [0..255]"
            )
        matrix_scaled_8u = matrix_scaled.clip(0, 255).astype(np.uint8)
        matrix_vis = cv2.applyColorMap(matrix_scaled_8u, self.cmap)
        matrix_vis[mask_bg] = image_target_bgr[y : y + h, x : x + w, :][mask_bg]
        image_target_bgr[y : y + h, x : x + w, :] = (
            image_target_bgr[y : y + h, x : x + w, :] * (1.0 - self.alpha) + matrix_vis * self.alpha
        )
        return image_target_bgr.astype(np.uint8)

    def _resize(self, mask, matrix, w, h):
        if (w != mask.shape[1]) or (h != mask.shape[0]):
            mask = cv2.resize(mask, (w, h), self.interp_method_mask)
        if (w != matrix.shape[1]) or (h != matrix.shape[0]):
            matrix = cv2.resize(matrix, (w, h), self.interp_method_matrix)
        return mask, matrix

    def _check_image(self, image_rgb):
        assert len(image_rgb.shape) == 3
        assert image_rgb.shape[2] == 3
        assert image_rgb.dtype == np.uint8

    def _check_mask_matrix(self, mask, matrix):
        assert len(matrix.shape) == 2
        assert len(mask.shape) == 2
        assert mask.dtype == np.uint8

class DensePoseResultsVisualizer:
    def visualize(
        self,
        image_bgr: Image,
        results,
    ) -> Image:
        context = self.create_visualization_context(image_bgr)
        for i, result in enumerate(results):
            boxes_xywh, labels, uv = result
            iuv_array = torch.cat(
                (labels[None].type(torch.float32), uv * 255.0)
            ).type(torch.uint8)
            self.visualize_iuv_arr(context, iuv_array.cpu().numpy(), boxes_xywh)
        image_bgr = self.context_to_image_bgr(context)
        return image_bgr

    def create_visualization_context(self, image_bgr: Image):
        return image_bgr

    def visualize_iuv_arr(self, context, iuv_arr: np.ndarray, bbox_xywh) -> None:
        pass

    def context_to_image_bgr(self, context):
        return context

    def get_image_bgr_from_context(self, context):
        return context

class DensePoseMaskedColormapResultsVisualizer(DensePoseResultsVisualizer):
    def __init__(
        self,
        data_extractor,
        segm_extractor,
        inplace=True,
        cmap=cv2.COLORMAP_PARULA,
        alpha=0.7,
        val_scale=1.0,
        **kwargs,
    ):
        self.mask_visualizer = MatrixVisualizer(
            inplace=inplace, cmap=cmap, val_scale=val_scale, alpha=alpha
        )
        self.data_extractor = data_extractor
        self.segm_extractor = segm_extractor

    def context_to_image_bgr(self, context):
        return context

    def visualize_iuv_arr(self, context, iuv_arr: np.ndarray, bbox_xywh) -> None:
        image_bgr = self.get_image_bgr_from_context(context)
        matrix = self.data_extractor(iuv_arr)
        segm = self.segm_extractor(iuv_arr)
        mask = np.zeros(matrix.shape, dtype=np.uint8)
        mask[segm > 0] = 1
        image_bgr = self.mask_visualizer.visualize(image_bgr, mask, matrix, bbox_xywh)


def _extract_i_from_iuvarr(iuv_arr):
    return iuv_arr[0, :, :]


def _extract_u_from_iuvarr(iuv_arr):
    return iuv_arr[1, :, :]


def _extract_v_from_iuvarr(iuv_arr):
    return iuv_arr[2, :, :]

def make_int_box(box: torch.Tensor) -> IntTupleBox:
    int_box = [0, 0, 0, 0]
    int_box[0], int_box[1], int_box[2], int_box[3] = tuple(box.long().tolist())
    return int_box[0], int_box[1], int_box[2], int_box[3]

def densepose_chart_predictor_output_to_result_with_confidences(
    boxes: Boxes,
    coarse_segm,
    fine_segm,
    u, v

):
    boxes_xyxy_abs = boxes.clone()
    boxes_xywh_abs = BoxMode.convert(boxes_xyxy_abs, BoxMode.XYXY_ABS, BoxMode.XYWH_ABS)
    box_xywh = make_int_box(boxes_xywh_abs[0])

    labels = resample_fine_and_coarse_segm_tensors_to_bbox(fine_segm, coarse_segm, box_xywh).squeeze(0)
    uv = resample_uv_tensors_to_bbox(u, v, labels, box_xywh)
    confidences = []
    return box_xywh, labels, uv

def resample_fine_and_coarse_segm_tensors_to_bbox(
    fine_segm: torch.Tensor, coarse_segm: torch.Tensor, box_xywh_abs: IntTupleBox
):
    """
    Resample fine and coarse segmentation tensors to the given
    bounding box and derive labels for each pixel of the bounding box

    Args:
        fine_segm: float tensor of shape [1, C, Hout, Wout]
        coarse_segm: float tensor of shape [1, K, Hout, Wout]
        box_xywh_abs (tuple of 4 int): bounding box given by its upper-left
            corner coordinates, width (W) and height (H)
    Return:
        Labels for each pixel of the bounding box, a long tensor of size [1, H, W]
    """
    x, y, w, h = box_xywh_abs
    w = max(int(w), 1)
    h = max(int(h), 1)
    # coarse segmentation
    coarse_segm_bbox = F.interpolate(
        coarse_segm,
        (h, w),
        mode="bilinear",
        align_corners=False,
    ).argmax(dim=1)
    # combined coarse and fine segmentation
    labels = (
        F.interpolate(fine_segm, (h, w), mode="bilinear", align_corners=False).argmax(dim=1)
        * (coarse_segm_bbox > 0).long()
    )
    return labels

def resample_uv_tensors_to_bbox(
    u: torch.Tensor,
    v: torch.Tensor,
    labels: torch.Tensor,
    box_xywh_abs: IntTupleBox,
) -> torch.Tensor:
    """
    Resamples U and V coordinate estimates for the given bounding box

    Args:
        u (tensor [1, C, H, W] of float): U coordinates
        v (tensor [1, C, H, W] of float): V coordinates
        labels (tensor [H, W] of long): labels obtained by resampling segmentation
            outputs for the given bounding box
        box_xywh_abs (tuple of 4 int): bounding box that corresponds to predictor outputs
    Return:
       Resampled U and V coordinates - a tensor [2, H, W] of float
    """
    x, y, w, h = box_xywh_abs
    w = max(int(w), 1)
    h = max(int(h), 1)
    u_bbox = F.interpolate(u, (h, w), mode="bilinear", align_corners=False)
    v_bbox = F.interpolate(v, (h, w), mode="bilinear", align_corners=False)
    uv = torch.zeros([2, h, w], dtype=torch.float32, device=u.device)
    for part_id in range(1, u_bbox.size(1)):
        uv[0][labels == part_id] = u_bbox[0, part_id][labels == part_id]
        uv[1][labels == part_id] = v_bbox[0, part_id][labels == part_id]
    return uv




from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class StableDiffusion(GraphNode):
    inputs: dict[str, Any] | GraphNode | tuple[GraphNode, str] = Field(default={}, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.image.generate.StableDiffusion"



class WorkflowNode(GraphNode):
    inputs: dict[str, Any] | GraphNode | tuple[GraphNode, str] = Field(default={}, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.workflows.workflow_node.Workflow"





from enum import Enum
from pydantic import Field
from nodetool.metadata.types import VAE, ImageTensor, Latent, Mask
from nodetool.common.comfy_node import MAX_RESOLUTION
from nodetool.common.comfy_node import ComfyNode


class LatentCompositeMasked(ComfyNode):
    """
    The Latent Composite Masked node can be used to paste a masked latent into another.
    """

    destination: Latent = Field(default=Latent(), description="The destination latent.")
    source: Latent = Field(default=Latent(), description="The source latent.")
    x: int = Field(default=0, description="The x position.")
    y: int = Field(default=0, description="The y position.")
    resize_source: bool = Field(
        default=False, description="Whether to resize the source."
    )
    mask: Mask = Field(None, description="The mask to use.")

    @classmethod
    def return_type(cls):
        return {"latent": Latent}


class EmptyLatentImage(ComfyNode):
    """
    The Empty Latent Image node can be used to create a new set of empty latent images. These latents can then be used inside e.g. a text2image workflow by noising and denoising them with a sampler node.
    """

    width: int = Field(
        default=512,
        description="The width of the latent image to generate.",
        ge=16,
        le=MAX_RESOLUTION,
    )
    height: int = Field(
        default=512,
        description="The height of the latent image to generate.",
        ge=16,
        le=MAX_RESOLUTION,
    )
    batch_size: int = Field(
        default=1,
        description="The batch size of the latent image to generate.",
        ge=1,
        le=MAX_RESOLUTION,
    )

    @classmethod
    def return_type(cls):
        return {"latent": Latent}


class VAEEncode(ComfyNode):
    """
    The VAE Encode node can be used to encode pixel space images into latent space images, using the provided VAE.
    """

    pixels: ImageTensor = Field(
        default=ImageTensor(), description="The image to encode."
    )
    vae: VAE = Field(default=VAE(), description="The VAE to use.")

    @classmethod
    def get_title(cls):
        return "VAE Encode"

    @classmethod
    def return_type(cls):
        return {"latent": Latent}


class VAEEncodeTiled(ComfyNode):
    pixels: ImageTensor = Field(
        default=ImageTensor(), description="The image pixels to encode."
    )
    vae: VAE = Field(default=VAE(), description="The VAE to use for encoding.")
    tile_size: int = Field(
        default=512,
        description="The size of the tiles to encode.",
        ge=320,  # ge is 'greater than or equal to'
        le=4096,  # le is 'less than or equal to'
        multiple_of=64,
    )

    @classmethod
    def return_type(cls):
        return {"latent": Latent}


class VAEEncodeForInpaint(ComfyNode):
    pixels: ImageTensor = Field(
        default=ImageTensor(), description="The image pixels to encode for inpainting."
    )
    vae: VAE = Field(default=VAE(), description="The VAE to use for encoding.")
    mask: Mask = Field(default=Mask(), description="The mask to apply for inpainting.")
    grow_mask_by: int = Field(
        default=6,
        description="Amount to grow the mask by.",
        ge=0,  # ge is 'greater than or equal to'
        le=64,  # le is 'less than or equal to'
    )

    @classmethod
    def return_type(cls):
        return {"latent": Latent}


class VAEDecode(ComfyNode):
    """
    The VAE Decode node can be used to decode latent space images back into pixel space images, using the provided VAE.
    """

    samples: Latent = Field(
        default=Latent(), description="The latent samples to decode."
    )
    vae: VAE = Field(default=VAE(), description="The VAE to use.")

    @classmethod
    def get_title(cls):
        return "VAE Decode"

    @classmethod
    def return_type(cls):
        return {"image": ImageTensor}


class VAEDecodeTiled(ComfyNode):
    """
    The VAE Decode node can be used to decode latent space images back into pixel space images, using the provided VAE.
    The tiled version of the node is useful for decoding large images that would otherwise exceed the memory limits of the system.
    """

    samples: Latent = Field(
        default=Latent(), description="The latent samples to decode."
    )
    vae: VAE = Field(default=VAE(), description="The VAE to use for decoding.")
    tile_size: int = Field(
        default=512,
        description="The size of the tiles to decode.",
        ge=320,  # ge is 'greater than or equal to'
        le=4096,  # le is 'less than or equal to'
        multiple_of=64,
    )

    @classmethod
    def return_type(cls):
        return {"image": ImageTensor}


# class SaveLatent(ComfyNode):
#     samples: Latent = Field(default=Latent(), description="The latent samples to save.")
#     filename_prefix: str = Field(
#         default="ComfyUI",
#         description="The prefix for the filename where the latents will be saved.",
#     )

# prompt: PromptRef = Field(
#     default=PromptRef(),
#     description="A hidden input for the prompt information.",
#     include_in_schema=False,
# )
# extra_pnginfo: ExtraPNGInfoRef = Field(
#     default=ExtraPNGInfoRef(),
#     description="Additional PNG information to be saved with the latent.",
#     include_in_schema=False,
# )j


class UpScaleMethod(str, Enum):
    NEAREST_EXACT = "nearest_exact"
    BILINEAR = "bilinear"
    AREA = "area"
    BICUBIC = "bicubic"
    BISLERP = "bislerp"


class CropMethod(str, Enum):
    DISABLED = "disabled"
    CENTER = "center"


# class LoadLatent(ComfyNode):
#     latent: Latent = Field(description="The latent file to load.")

#     @classmethod
#     def return_type(cls):
#         return {"latent": Latent}


class LatentUpscale(ComfyNode):
    """
    The Upscale Latent node can be used to resize latent images.
    """

    samples: Latent = Field(
        default=Latent(), description="The latent samples to upscale."
    )
    upscale_method: UpScaleMethod = Field(
        default=UpScaleMethod.NEAREST_EXACT,
        description="The method to use for upscaling.",
    )
    width: int = Field(
        default=512,
        description="The target width after upscaling.",
        ge=0,
        le=MAX_RESOLUTION,
        multiple_of=8,
    )
    height: int = Field(
        default=512,
        description="The target height after upscaling.",
        ge=0,
        le=MAX_RESOLUTION,
        multiple_of=8,
    )
    crop: CropMethod = Field(
        default=CropMethod.DISABLED, description="The method to use for cropping."
    )

    @classmethod
    def get_title(cls):
        return "Upscale Latent"

    @classmethod
    def return_type(cls):
        return {"latent": Latent}


class LatentUpscaleBy(ComfyNode):
    """
    The Upscale Latent node can be used to resize latent images.
    """

    samples: Latent = Field(
        default=Latent(), description="The latent samples to upscale."
    )
    upscale_method: UpScaleMethod = Field(
        default=UpScaleMethod.NEAREST_EXACT,
        description="The method to use for upscaling.",
    )
    scale_by: float = Field(
        default=1.5,
        description="The factor by which to scale.",
        ge=0.01,
        le=8.0,
        multiple_of=0.01,
    )

    @classmethod
    def get_title(cls):
        return "Upscale Latent By"

    @classmethod
    def return_type(cls):
        return {"latent": Latent}


class LatentComposite(ComfyNode):
    """
    The Latent Composite node can be used to paste one latent into another.
    """

    samples_to: Latent = Field(default=Latent(), description="The first latent sample.")
    samples_from: Latent = Field(
        default=Latent(), description="The second latent sample."
    )
    x: int = Field(default=0, description="The x-coordinate for compositing.")
    y: int = Field(default=0, description="The y-coordinate for compositing.")
    feather: int = Field(
        default=0, description="The feather amount for compositing edges."
    )

    @classmethod
    def return_type(cls):
        return {"latent": Latent}


class LatentBlend(ComfyNode):
    samples1: Latent = Field(
        default=Latent(), description="The first set of latent samples."
    )
    samples2: Latent = Field(
        default=Latent(), description="The second set of latent samples."
    )
    blend_factor: float = Field(
        default=0.5, description="The blend factor between samples."
    )

    @classmethod
    def return_type(cls):
        return {"latent": Latent}



from pydantic import BaseModel, Field
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode
from nodetool.common.replicate_node import ReplicateNode
from enum import Enum


class Rvc_model(str, Enum):
    SQUIDWARD = 'Squidward'
    MRKRABS = 'MrKrabs'
    PLANKTON = 'Plankton'
    DRAKE = 'Drake'
    VADER = 'Vader'
    TRUMP = 'Trump'
    BIDEN = 'Biden'
    OBAMA = 'Obama'
    GUITAR = 'Guitar'
    VOILIN = 'Voilin'
    CUSTOM = 'CUSTOM'
class Pitch_change(str, Enum):
    NO_CHANGE = 'no-change'
    MALE_TO_FEMALE = 'male-to-female'
    FEMALE_TO_MALE = 'female-to-male'
class Output_format(str, Enum):
    MP3 = 'mp3'
    WAV = 'wav'
class Pitch_detection_algorithm(str, Enum):
    RMVPE = 'rmvpe'
    MANGIO_CREPE = 'mangio-crepe'


class RealisticVoiceCloning(ReplicateNode):
    """Create song covers with any RVC v2 trained AI voice from audio files."""

    @classmethod
    def replicate_model_id(cls): return 'zsxkib/realistic-voice-cloning:0a9c7c558af4c0f20667c1bd1260ce32a2879944a0b9e44e1398660c077b1550'
    @classmethod
    def get_hardware(cls): return 'Nvidia T4 GPU'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/7a110507-8a38-4fcb-8826-61aec80309cb/Out_0_1024x1024.png', 'created_at': '2023-11-09T16:32:42.062982Z', 'description': 'Create song covers with any RVC v2 trained AI voice from audio files.', 'github_url': 'https://github.com/zsxkib/AICoverGen.git', 'license_url': 'https://github.com/SociallyIneptWeeb/AICoverGen/blob/main/LICENSE', 'name': 'realistic-voice-cloning', 'owner': 'zsxkib', 'paper_url': None, 'run_count': 211293, 'url': 'https://replicate.com/zsxkib/realistic-voice-cloning', 'visibility': 'public', 'hardware': 'Nvidia T4 GPU'}
    @classmethod
    def return_type(cls): return AudioRef


    protect: float = Field(title='Protect', description="Control how much of the original vocals' breath and voiceless consonants to leave in the AI vocals. Set 0.5 to disable.", ge=0.0, le=0.5, default=0.33)
    rvc_model: Rvc_model = Field(description="RVC model for a specific voice. If using a custom model, this should match the name of the downloaded model. If a 'custom_rvc_model_download_url' is provided, this will be automatically set to the name of the downloaded model.", default='Squidward')
    index_rate: float = Field(title='Index Rate', description="Control how much of the AI's accent to leave in the vocals.", ge=0.0, le=1.0, default=0.5)
    song_input: AudioRef = Field(default=AudioRef(), description='Upload your audio file here.')
    reverb_size: float = Field(title='Reverb Size', description='The larger the room, the longer the reverb time.', ge=0.0, le=1.0, default=0.15)
    pitch_change: Pitch_change = Field(description='Adjust pitch of AI vocals. Options: `no-change`, `male-to-female`, `female-to-male`.', default='no-change')
    rms_mix_rate: float = Field(title='Rms Mix Rate', description="Control how much to use the original vocal's loudness (0) or a fixed loudness (1).", ge=0.0, le=1.0, default=0.25)
    filter_radius: int = Field(title='Filter Radius', description='If >=3: apply median filtering median filtering to the harvested pitch results.', ge=0.0, le=7.0, default=3)
    output_format: Output_format = Field(description='wav for best quality and large file size, mp3 for decent quality and small file size.', default='mp3')
    reverb_damping: float = Field(title='Reverb Damping', description='Absorption of high frequencies in the reverb.', ge=0.0, le=1.0, default=0.7)
    reverb_dryness: float = Field(title='Reverb Dryness', description='Level of AI vocals without reverb.', ge=0.0, le=1.0, default=0.8)
    reverb_wetness: float = Field(title='Reverb Wetness', description='Level of AI vocals with reverb.', ge=0.0, le=1.0, default=0.2)
    crepe_hop_length: int = Field(title='Crepe Hop Length', description='When `pitch_detection_algo` is set to `mangio-crepe`, this controls how often it checks for pitch changes in milliseconds. Lower values lead to longer conversions and higher risk of voice cracks, but better pitch accuracy.', default=128)
    pitch_change_all: float = Field(title='Pitch Change All', description='Change pitch/key of background music, backup vocals and AI vocals in semitones. Reduces sound quality slightly.', default=0)
    main_vocals_volume_change: float = Field(title='Main Vocals Volume Change', description='Control volume of main AI vocals. Use -3 to decrease the volume by 3 decibels, or 3 to increase the volume by 3 decibels.', default=0)
    pitch_detection_algorithm: Pitch_detection_algorithm = Field(description='Best option is rmvpe (clarity in vocals), then mangio-crepe (smoother vocals).', default='rmvpe')
    instrumental_volume_change: float = Field(title='Instrumental Volume Change', description='Control volume of the background music/instrumentals.', default=0)
    backup_vocals_volume_change: float = Field(title='Backup Vocals Volume Change', description='Control volume of backup AI vocals.', default=0)
    custom_rvc_model_download_url: str | None = Field(title='Custom Rvc Model Download Url', description="URL to download a custom RVC model. If provided, the model will be downloaded (if it doesn't already exist) and used for prediction, regardless of the 'rvc_model' value.", default=None)

class Preset(str, Enum):
    ULTRA_FAST = 'ultra_fast'
    FAST = 'fast'
    STANDARD = 'standard'
    HIGH_QUALITY = 'high_quality'
class Voice_a(str, Enum):
    ANGIE = 'angie'
    COND_LATENT_EXAMPLE = 'cond_latent_example'
    DENIRO = 'deniro'
    FREEMAN = 'freeman'
    HALLE = 'halle'
    LJ = 'lj'
    MYSELF = 'myself'
    PAT2 = 'pat2'
    SNAKES = 'snakes'
    TOM = 'tom'
    TRAIN_DAWS = 'train_daws'
    TRAIN_DREAMS = 'train_dreams'
    TRAIN_GRACE = 'train_grace'
    TRAIN_LESCAULT = 'train_lescault'
    WEAVER = 'weaver'
    APPLEJACK = 'applejack'
    DANIEL = 'daniel'
    EMMA = 'emma'
    GERALT = 'geralt'
    JLAW = 'jlaw'
    MOL = 'mol'
    PAT = 'pat'
    RAINBOW = 'rainbow'
    TIM_REYNOLDS = 'tim_reynolds'
    TRAIN_ATKINS = 'train_atkins'
    TRAIN_DOTRICE = 'train_dotrice'
    TRAIN_EMPIRE = 'train_empire'
    TRAIN_KENNARD = 'train_kennard'
    TRAIN_MOUSE = 'train_mouse'
    WILLIAM = 'william'
    RANDOM = 'random'
    CUSTOM_VOICE = 'custom_voice'
    DISABLED = 'disabled'
class Voice_b(str, Enum):
    ANGIE = 'angie'
    COND_LATENT_EXAMPLE = 'cond_latent_example'
    DENIRO = 'deniro'
    FREEMAN = 'freeman'
    HALLE = 'halle'
    LJ = 'lj'
    MYSELF = 'myself'
    PAT2 = 'pat2'
    SNAKES = 'snakes'
    TOM = 'tom'
    TRAIN_DAWS = 'train_daws'
    TRAIN_DREAMS = 'train_dreams'
    TRAIN_GRACE = 'train_grace'
    TRAIN_LESCAULT = 'train_lescault'
    WEAVER = 'weaver'
    APPLEJACK = 'applejack'
    DANIEL = 'daniel'
    EMMA = 'emma'
    GERALT = 'geralt'
    JLAW = 'jlaw'
    MOL = 'mol'
    PAT = 'pat'
    RAINBOW = 'rainbow'
    TIM_REYNOLDS = 'tim_reynolds'
    TRAIN_ATKINS = 'train_atkins'
    TRAIN_DOTRICE = 'train_dotrice'
    TRAIN_EMPIRE = 'train_empire'
    TRAIN_KENNARD = 'train_kennard'
    TRAIN_MOUSE = 'train_mouse'
    WILLIAM = 'william'
    RANDOM = 'random'
    CUSTOM_VOICE = 'custom_voice'
    DISABLED = 'disabled'
class Voice_c(str, Enum):
    ANGIE = 'angie'
    COND_LATENT_EXAMPLE = 'cond_latent_example'
    DENIRO = 'deniro'
    FREEMAN = 'freeman'
    HALLE = 'halle'
    LJ = 'lj'
    MYSELF = 'myself'
    PAT2 = 'pat2'
    SNAKES = 'snakes'
    TOM = 'tom'
    TRAIN_DAWS = 'train_daws'
    TRAIN_DREAMS = 'train_dreams'
    TRAIN_GRACE = 'train_grace'
    TRAIN_LESCAULT = 'train_lescault'
    WEAVER = 'weaver'
    APPLEJACK = 'applejack'
    DANIEL = 'daniel'
    EMMA = 'emma'
    GERALT = 'geralt'
    JLAW = 'jlaw'
    MOL = 'mol'
    PAT = 'pat'
    RAINBOW = 'rainbow'
    TIM_REYNOLDS = 'tim_reynolds'
    TRAIN_ATKINS = 'train_atkins'
    TRAIN_DOTRICE = 'train_dotrice'
    TRAIN_EMPIRE = 'train_empire'
    TRAIN_KENNARD = 'train_kennard'
    TRAIN_MOUSE = 'train_mouse'
    WILLIAM = 'william'
    RANDOM = 'random'
    CUSTOM_VOICE = 'custom_voice'
    DISABLED = 'disabled'


class TortoiseTTS(ReplicateNode):
    """Generate speech from text, clone voices from mp3 files. From James Betker AKA "neonbjb"."""

    @classmethod
    def replicate_model_id(cls): return 'afiaka87/tortoise-tts:e9658de4b325863c4fcdc12d94bb7c9b54cbfe351b7ca1b36860008172b91c71'
    @classmethod
    def get_hardware(cls): return 'Nvidia T4 GPU'
    @classmethod
    def model_info(cls): return {'cover_image_url': None, 'created_at': '2022-08-02T02:01:54.555794Z', 'description': 'Generate speech from text, clone voices from mp3 files. From James Betker AKA "neonbjb".', 'github_url': 'https://github.com/afiaka87/tortoise-tts', 'license_url': 'https://github.com/afiaka87/tortoise-tts/blob/main/LICENSE', 'name': 'tortoise-tts', 'owner': 'afiaka87', 'paper_url': 'https://github.com/neonbjb/tortoise-tts', 'run_count': 158226, 'url': 'https://replicate.com/afiaka87/tortoise-tts', 'visibility': 'public', 'hardware': 'Nvidia T4 GPU'}
    @classmethod
    def return_type(cls): return AudioRef


    seed: int = Field(title='Seed', description='Random seed which can be used to reproduce results.', default=0)
    text: str = Field(title='Text', description='Text to speak.', default='The expressiveness of autoregressive transformers is literally nuts! I absolutely adore them.')
    preset: Preset = Field(description='Which voice preset to use. See the documentation for more information.', default='fast')
    voice_a: Voice_a = Field(description='Selects the voice to use for generation. Use `random` to select a random voice. Use `custom_voice` to use a custom voice.', default='random')
    voice_b: Voice_b = Field(description='(Optional) Create new voice from averaging the latents for `voice_a`, `voice_b` and `voice_c`. Use `disabled` to disable voice mixing.', default='disabled')
    voice_c: Voice_c = Field(description='(Optional) Create new voice from averaging the latents for `voice_a`, `voice_b` and `voice_c`. Use `disabled` to disable voice mixing.', default='disabled')
    cvvp_amount: float = Field(title='Cvvp Amount', description='How much the CVVP model should influence the output. Increasing this can in some cases reduce the likelyhood of multiple speakers. Defaults to 0 (disabled)', ge=0.0, le=1.0, default=0)
    custom_voice: AudioRef = Field(default=AudioRef(), description='(Optional) Create a custom voice based on an mp3 file of a speaker. Audio should be at least 15 seconds, only contain one speaker, and be in mp3 format. Overrides the `voice_a` input.')



class StyleTTS2(ReplicateNode):
    """Generates speech from text"""

    @classmethod
    def replicate_model_id(cls): return 'adirik/styletts2:989cb5ea6d2401314eb30685740cb9f6fd1c9001b8940659b406f952837ab5ac'
    @classmethod
    def get_hardware(cls): return 'Nvidia T4 GPU'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/b3443880-411e-4b5f-b9f9-3db28c59b578/out-0.png', 'created_at': '2023-11-20T19:22:15.416691Z', 'description': 'Generates speech from text', 'github_url': 'https://github.com/yl4579/StyleTTS2', 'license_url': 'https://github.com/yl4579/StyleTTS2/blob/main/LICENSE', 'name': 'styletts2', 'owner': 'adirik', 'paper_url': 'https://arxiv.org/abs/2306.07691', 'run_count': 102550, 'url': 'https://replicate.com/adirik/styletts2', 'visibility': 'public', 'hardware': 'Nvidia T4 GPU'}
    @classmethod
    def return_type(cls): return AudioRef


    beta: float = Field(title='Beta', description='Only used for long text inputs or in case of reference speaker,             determines the prosody of the speaker. Use lower values to sample style based             on previous or reference speech instead of text.', ge=0.0, le=1.0, default=0.7)
    seed: int = Field(title='Seed', description='Seed for reproducibility', default=0)
    text: str | None = Field(title='Text', description='Text to convert to speech', default=None)
    alpha: float = Field(title='Alpha', description='Only used for long text inputs or in case of reference speaker,             determines the timbre of the speaker. Use lower values to sample style based             on previous or reference speech instead of text.', ge=0.0, le=1.0, default=0.3)
    weights: str | None = Field(title='Weights', description='Replicate weights url for inference with model that is fine-tuned on new speakers.            If provided, a reference speech must also be provided.             If not provided, the default model will be used.', default=None)
    reference: AudioRef = Field(default=AudioRef(), description='Reference speech to copy style from')
    diffusion_steps: int = Field(title='Diffusion Steps', description='Number of diffusion steps', ge=0.0, le=50.0, default=10)
    embedding_scale: float = Field(title='Embedding Scale', description='Embedding scale, use higher values for pronounced emotion', ge=0.0, le=5.0, default=1)

class Seed_image_id(str, Enum):
    AGILE = 'agile'
    MARIM = 'marim'
    MASK_BEAT_LINES_80 = 'mask_beat_lines_80'
    MASK_GRADIENT_DARK = 'mask_gradient_dark'
    MASK_GRADIENT_TOP_70 = 'mask_gradient_top_70'
    MASK_GRAIDENT_TOP_FIFTH_75 = 'mask_graident_top_fifth_75'
    MASK_TOP_THIRD_75 = 'mask_top_third_75'
    MASK_TOP_THIRD_95 = 'mask_top_third_95'
    MOTORWAY = 'motorway'
    OG_BEAT = 'og_beat'
    VIBES = 'vibes'


class Riffusion(ReplicateNode):
    """Stable diffusion for real-time music generation"""

    @classmethod
    def replicate_model_id(cls): return 'riffusion/riffusion:8cf61ea6c56afd61d8f5b9ffd14d7c216c0a93844ce2d82ac1c9ecc9c7f24e05'
    @classmethod
    def get_hardware(cls): return 'Nvidia T4 GPU'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/4f67b4e2-4df3-4ce9-9b80-00bbf5a186d8/riffusion.gif', 'created_at': '2022-12-16T07:31:34.983811Z', 'description': 'Stable diffusion for real-time music generation', 'github_url': 'https://github.com/riffusion/riffusion', 'license_url': 'https://github.com/riffusion/riffusion/blob/main/LICENSE', 'name': 'riffusion', 'owner': 'riffusion', 'paper_url': 'https://www.riffusion.com/about', 'run_count': 918272, 'url': 'https://replicate.com/riffusion/riffusion', 'visibility': 'public', 'hardware': 'Nvidia T4 GPU'}
    @classmethod
    def return_type(cls): return AudioRef
    def output_key(self): return 'audio'


    alpha: float = Field(title='Alpha', description='Interpolation alpha if using two prompts. A value of 0 uses prompt_a fully, a value of 1 uses prompt_b fully', ge=0.0, le=1.0, default=0.5)
    prompt_a: str = Field(title='Prompt A', description='The prompt for your audio', default='funky synth solo')
    prompt_b: str | None = Field(title='Prompt B', description='The second prompt to interpolate with the first, leave blank if no interpolation', default=None)
    denoising: float = Field(title='Denoising', description='How much to transform input spectrogram', ge=0.0, le=1.0, default=0.75)
    seed_image_id: Seed_image_id = Field(description='Seed spectrogram to use', default='vibes')
    num_inference_steps: int = Field(title='Num Inference Steps', description='Number of steps to run the diffusion model', ge=1.0, default=50)

class History_prompt(str, Enum):
    ANNOUNCER = 'announcer'
    DE_SPEAKER_0 = 'de_speaker_0'
    DE_SPEAKER_1 = 'de_speaker_1'
    DE_SPEAKER_2 = 'de_speaker_2'
    DE_SPEAKER_3 = 'de_speaker_3'
    DE_SPEAKER_4 = 'de_speaker_4'
    DE_SPEAKER_5 = 'de_speaker_5'
    DE_SPEAKER_6 = 'de_speaker_6'
    DE_SPEAKER_7 = 'de_speaker_7'
    DE_SPEAKER_8 = 'de_speaker_8'
    DE_SPEAKER_9 = 'de_speaker_9'
    EN_SPEAKER_0 = 'en_speaker_0'
    EN_SPEAKER_1 = 'en_speaker_1'
    EN_SPEAKER_2 = 'en_speaker_2'
    EN_SPEAKER_3 = 'en_speaker_3'
    EN_SPEAKER_4 = 'en_speaker_4'
    EN_SPEAKER_5 = 'en_speaker_5'
    EN_SPEAKER_6 = 'en_speaker_6'
    EN_SPEAKER_7 = 'en_speaker_7'
    EN_SPEAKER_8 = 'en_speaker_8'
    EN_SPEAKER_9 = 'en_speaker_9'
    ES_SPEAKER_0 = 'es_speaker_0'
    ES_SPEAKER_1 = 'es_speaker_1'
    ES_SPEAKER_2 = 'es_speaker_2'
    ES_SPEAKER_3 = 'es_speaker_3'
    ES_SPEAKER_4 = 'es_speaker_4'
    ES_SPEAKER_5 = 'es_speaker_5'
    ES_SPEAKER_6 = 'es_speaker_6'
    ES_SPEAKER_7 = 'es_speaker_7'
    ES_SPEAKER_8 = 'es_speaker_8'
    ES_SPEAKER_9 = 'es_speaker_9'
    FR_SPEAKER_0 = 'fr_speaker_0'
    FR_SPEAKER_1 = 'fr_speaker_1'
    FR_SPEAKER_2 = 'fr_speaker_2'
    FR_SPEAKER_3 = 'fr_speaker_3'
    FR_SPEAKER_4 = 'fr_speaker_4'
    FR_SPEAKER_5 = 'fr_speaker_5'
    FR_SPEAKER_6 = 'fr_speaker_6'
    FR_SPEAKER_7 = 'fr_speaker_7'
    FR_SPEAKER_8 = 'fr_speaker_8'
    FR_SPEAKER_9 = 'fr_speaker_9'
    HI_SPEAKER_0 = 'hi_speaker_0'
    HI_SPEAKER_1 = 'hi_speaker_1'
    HI_SPEAKER_2 = 'hi_speaker_2'
    HI_SPEAKER_3 = 'hi_speaker_3'
    HI_SPEAKER_4 = 'hi_speaker_4'
    HI_SPEAKER_5 = 'hi_speaker_5'
    HI_SPEAKER_6 = 'hi_speaker_6'
    HI_SPEAKER_7 = 'hi_speaker_7'
    HI_SPEAKER_8 = 'hi_speaker_8'
    HI_SPEAKER_9 = 'hi_speaker_9'
    IT_SPEAKER_0 = 'it_speaker_0'
    IT_SPEAKER_1 = 'it_speaker_1'
    IT_SPEAKER_2 = 'it_speaker_2'
    IT_SPEAKER_3 = 'it_speaker_3'
    IT_SPEAKER_4 = 'it_speaker_4'
    IT_SPEAKER_5 = 'it_speaker_5'
    IT_SPEAKER_6 = 'it_speaker_6'
    IT_SPEAKER_7 = 'it_speaker_7'
    IT_SPEAKER_8 = 'it_speaker_8'
    IT_SPEAKER_9 = 'it_speaker_9'
    JA_SPEAKER_0 = 'ja_speaker_0'
    JA_SPEAKER_1 = 'ja_speaker_1'
    JA_SPEAKER_2 = 'ja_speaker_2'
    JA_SPEAKER_3 = 'ja_speaker_3'
    JA_SPEAKER_4 = 'ja_speaker_4'
    JA_SPEAKER_5 = 'ja_speaker_5'
    JA_SPEAKER_6 = 'ja_speaker_6'
    JA_SPEAKER_7 = 'ja_speaker_7'
    JA_SPEAKER_8 = 'ja_speaker_8'
    JA_SPEAKER_9 = 'ja_speaker_9'
    KO_SPEAKER_0 = 'ko_speaker_0'
    KO_SPEAKER_1 = 'ko_speaker_1'
    KO_SPEAKER_2 = 'ko_speaker_2'
    KO_SPEAKER_3 = 'ko_speaker_3'
    KO_SPEAKER_4 = 'ko_speaker_4'
    KO_SPEAKER_5 = 'ko_speaker_5'
    KO_SPEAKER_6 = 'ko_speaker_6'
    KO_SPEAKER_7 = 'ko_speaker_7'
    KO_SPEAKER_8 = 'ko_speaker_8'
    KO_SPEAKER_9 = 'ko_speaker_9'
    PL_SPEAKER_0 = 'pl_speaker_0'
    PL_SPEAKER_1 = 'pl_speaker_1'
    PL_SPEAKER_2 = 'pl_speaker_2'
    PL_SPEAKER_3 = 'pl_speaker_3'
    PL_SPEAKER_4 = 'pl_speaker_4'
    PL_SPEAKER_5 = 'pl_speaker_5'
    PL_SPEAKER_6 = 'pl_speaker_6'
    PL_SPEAKER_7 = 'pl_speaker_7'
    PL_SPEAKER_8 = 'pl_speaker_8'
    PL_SPEAKER_9 = 'pl_speaker_9'
    PT_SPEAKER_0 = 'pt_speaker_0'
    PT_SPEAKER_1 = 'pt_speaker_1'
    PT_SPEAKER_2 = 'pt_speaker_2'
    PT_SPEAKER_3 = 'pt_speaker_3'
    PT_SPEAKER_4 = 'pt_speaker_4'
    PT_SPEAKER_5 = 'pt_speaker_5'
    PT_SPEAKER_6 = 'pt_speaker_6'
    PT_SPEAKER_7 = 'pt_speaker_7'
    PT_SPEAKER_8 = 'pt_speaker_8'
    PT_SPEAKER_9 = 'pt_speaker_9'
    RU_SPEAKER_0 = 'ru_speaker_0'
    RU_SPEAKER_1 = 'ru_speaker_1'
    RU_SPEAKER_2 = 'ru_speaker_2'
    RU_SPEAKER_3 = 'ru_speaker_3'
    RU_SPEAKER_4 = 'ru_speaker_4'
    RU_SPEAKER_5 = 'ru_speaker_5'
    RU_SPEAKER_6 = 'ru_speaker_6'
    RU_SPEAKER_7 = 'ru_speaker_7'
    RU_SPEAKER_8 = 'ru_speaker_8'
    RU_SPEAKER_9 = 'ru_speaker_9'
    TR_SPEAKER_0 = 'tr_speaker_0'
    TR_SPEAKER_1 = 'tr_speaker_1'
    TR_SPEAKER_2 = 'tr_speaker_2'
    TR_SPEAKER_3 = 'tr_speaker_3'
    TR_SPEAKER_4 = 'tr_speaker_4'
    TR_SPEAKER_5 = 'tr_speaker_5'
    TR_SPEAKER_6 = 'tr_speaker_6'
    TR_SPEAKER_7 = 'tr_speaker_7'
    TR_SPEAKER_8 = 'tr_speaker_8'
    TR_SPEAKER_9 = 'tr_speaker_9'
    ZH_SPEAKER_0 = 'zh_speaker_0'
    ZH_SPEAKER_1 = 'zh_speaker_1'
    ZH_SPEAKER_2 = 'zh_speaker_2'
    ZH_SPEAKER_3 = 'zh_speaker_3'
    ZH_SPEAKER_4 = 'zh_speaker_4'
    ZH_SPEAKER_5 = 'zh_speaker_5'
    ZH_SPEAKER_6 = 'zh_speaker_6'
    ZH_SPEAKER_7 = 'zh_speaker_7'
    ZH_SPEAKER_8 = 'zh_speaker_8'
    ZH_SPEAKER_9 = 'zh_speaker_9'


class Bark(ReplicateNode):
    """🔊 Text-Prompted Generative Audio Model"""

    @classmethod
    def replicate_model_id(cls): return 'suno-ai/bark:b76242b40d67c76ab6742e987628a2a9ac019e11d56ab96c4e91ce03b79b2787'
    @classmethod
    def get_hardware(cls): return 'Nvidia T4 GPU'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/9fbbe7c2-beb0-4caa-bd59-647dd7eeee6b/fofr_dog.jpg', 'created_at': '2023-04-25T20:56:39.740033Z', 'description': '🔊 Text-Prompted Generative Audio Model', 'github_url': 'https://github.com/chenxwh/bark', 'license_url': 'https://github.com/suno-ai/bark/blob/main/LICENSE', 'name': 'bark', 'owner': 'suno-ai', 'paper_url': None, 'run_count': 234192, 'url': 'https://replicate.com/suno-ai/bark', 'visibility': 'public', 'hardware': 'Nvidia T4 GPU'}
    @classmethod
    def return_type(cls): return AudioRef
    def output_key(self): return 'audio_out'


    prompt: str = Field(title='Prompt', description='Input prompt', default='Hello, my name is Suno. And, uh — and I like pizza. [laughs] But I also have other interests such as playing tic tac toe.')
    text_temp: float = Field(title='Text Temp', description='generation temperature (1.0 more diverse, 0.0 more conservative)', default=0.7)
    output_full: bool = Field(title='Output Full', description='return full generation as a .npz file to be used as a history prompt', default=False)
    waveform_temp: float = Field(title='Waveform Temp', description='generation temperature (1.0 more diverse, 0.0 more conservative)', default=0.7)
    history_prompt: History_prompt = Field(description='history choice for audio cloning, choose from the list', default=None)
    custom_history_prompt: str | None = Field(title='Custom History Prompt', description='Provide your own .npz file with history choice for audio cloning, this will override the previous history_prompt setting', default=None)

class Model_version(str, Enum):
    STEREO_MELODY_LARGE = 'stereo-melody-large'
    STEREO_LARGE = 'stereo-large'
    MELODY_LARGE = 'melody-large'
    LARGE = 'large'
class Normalization_strategy(str, Enum):
    LOUDNESS = 'loudness'
    CLIP = 'clip'
    PEAK = 'peak'
    RMS = 'rms'


class MusicGen(ReplicateNode):
    """Generate music from a prompt or melody"""

    @classmethod
    def replicate_model_id(cls): return 'meta/musicgen:671ac645ce5e552cc63a54a2bbff63fcf798043055d2dac5fc9e36a837eedcfb'
    @classmethod
    def get_hardware(cls): return 'Nvidia A100 (40GB) GPU'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/e3d6b391-571b-4e24-8531-936a0fee9fba/musicgen.jpeg', 'created_at': '2023-06-12T19:22:05.525230Z', 'description': 'Generate music from a prompt or melody', 'github_url': 'https://github.com/replicate/cog-musicgen', 'license_url': 'https://github.com/facebookresearch/audiocraft/blob/main/LICENSE_weights', 'name': 'musicgen', 'owner': 'meta', 'paper_url': 'https://arxiv.org/abs/2306.05284', 'run_count': 1729400, 'url': 'https://replicate.com/meta/musicgen', 'visibility': 'public', 'hardware': 'Nvidia A100 (40GB) GPU'}
    @classmethod
    def return_type(cls): return AudioRef


    seed: int | None = Field(title='Seed', description='Seed for random number generator. If None or -1, a random seed will be used.', default=None)
    top_k: int = Field(title='Top K', description='Reduces sampling to the k most likely tokens.', default=250)
    top_p: float = Field(title='Top P', description='Reduces sampling to tokens with cumulative probability of p. When set to  `0` (default), top_k sampling is used.', default=0)
    prompt: str | None = Field(title='Prompt', description='A description of the music you want to generate.', default=None)
    duration: int = Field(title='Duration', description='Duration of the generated audio in seconds.', default=8)
    input_audio: str | None = Field(title='Input Audio', description="An audio file that will influence the generated music. If `continuation` is `True`, the generated music will be a continuation of the audio file. Otherwise, the generated music will mimic the audio file's melody.", default=None)
    temperature: float = Field(title='Temperature', description="Controls the 'conservativeness' of the sampling process. Higher temperature means more diversity.", default=1)
    continuation: bool = Field(title='Continuation', description="If `True`, generated music will continue from `input_audio`. Otherwise, generated music will mimic `input_audio`'s melody.", default=False)
    model_version: Model_version = Field(description='Model to use for generation', default='stereo-melody-large')
    output_format: Output_format = Field(description='Output format for generated audio.', default='wav')
    continuation_end: int | None = Field(title='Continuation End', description='End time of the audio file to use for continuation. If -1 or None, will default to the end of the audio clip.', ge=0.0, default=None)
    continuation_start: int = Field(title='Continuation Start', description='Start time of the audio file to use for continuation.', ge=0.0, default=0)
    multi_band_diffusion: bool = Field(title='Multi Band Diffusion', description='If `True`, the EnCodec tokens will be decoded with MultiBand Diffusion. Only works with non-stereo models.', default=False)
    normalization_strategy: Normalization_strategy = Field(description='Strategy for normalizing audio.', default='loudness')
    classifier_free_guidance: int = Field(title='Classifier Free Guidance', description='Increases the influence of inputs on the output. Higher values produce lower-varience outputs that adhere more closely to inputs.', default=3)


import psutil
import logging
from enum import Enum
from comfy.cli_args import args
import comfy.utils
import torch
import sys


class VRAMState(Enum):
    DISABLED = 0  # No vram present: no need to move models to vram
    NO_VRAM = 1  # Very low vram: enable all the options to save vram
    LOW_VRAM = 2
    NORMAL_VRAM = 3
    HIGH_VRAM = 4
    SHARED = 5  # No dedicated vram: memory shared between CPU and GPU but models still need to be moved between both.


class CPUState(Enum):
    GPU = 0
    CPU = 1
    MPS = 2


# Determine VRAM State
vram_state = VRAMState.NORMAL_VRAM
set_vram_to = VRAMState.NORMAL_VRAM
cpu_state = CPUState.GPU

total_vram = 0

lowvram_available = True
xpu_available = False

if args.deterministic:
    logging.info("Using deterministic algorithms for pytorch")
    torch.use_deterministic_algorithms(True, warn_only=True)

directml_enabled = False
if args.directml is not None:
    import torch_directml

    directml_enabled = True
    device_index = args.directml
    if device_index < 0:
        directml_device = torch_directml.device()
    else:
        directml_device = torch_directml.device(device_index)
    logging.info(
        "Using directml with device: {}".format(
            torch_directml.device_name(device_index)
        )
    )
    # torch_directml.disable_tiled_resources(True)
    lowvram_available = False  # TODO: need to find a way to get free memory in directml before this can be enabled by default.

try:
    import intel_extension_for_pytorch as ipex

    if torch.xpu.is_available():
        xpu_available = True
except:
    pass

try:
    if torch.backends.mps.is_available():
        cpu_state = CPUState.MPS
        import torch.mps
except:
    pass

if args.cpu:
    cpu_state = CPUState.CPU


def is_intel_xpu():
    global cpu_state
    global xpu_available
    if cpu_state == CPUState.GPU:
        if xpu_available:
            return True
    return False


def get_torch_device():
    global directml_enabled
    global cpu_state
    if directml_enabled:
        global directml_device
        return directml_device
    if cpu_state == CPUState.MPS:
        return torch.device("mps")
    if cpu_state == CPUState.CPU:
        return torch.device("cpu")
    else:
        if is_intel_xpu():
            return torch.device("xpu")
        else:
            return torch.device(torch.cuda.current_device())


def get_total_memory(dev=None, torch_total_too=False):
    global directml_enabled
    if dev is None:
        dev = get_torch_device()

    if hasattr(dev, "type") and (dev.type == "cpu" or dev.type == "mps"):
        mem_total = psutil.virtual_memory().total
        mem_total_torch = mem_total
    else:
        if directml_enabled:
            mem_total = 1024 * 1024 * 1024  # TODO
            mem_total_torch = mem_total
        elif is_intel_xpu():
            stats = torch.xpu.memory_stats(dev)
            mem_reserved = stats["reserved_bytes.all.current"]
            mem_total = torch.xpu.get_device_properties(dev).total_memory
            mem_total_torch = mem_reserved
        else:
            stats = torch.cuda.memory_stats(dev)
            mem_reserved = stats["reserved_bytes.all.current"]
            _, mem_total_cuda = torch.cuda.mem_get_info(dev)
            mem_total_torch = mem_reserved
            mem_total = mem_total_cuda

    if torch_total_too:
        return (mem_total, mem_total_torch)
    else:
        return mem_total


total_vram = get_total_memory(get_torch_device()) / (1024 * 1024)
total_ram = psutil.virtual_memory().total / (1024 * 1024)
logging.info(
    "Total VRAM {:0.0f} MB, total RAM {:0.0f} MB".format(total_vram, total_ram)
)
if not args.normalvram and not args.cpu:
    if lowvram_available and total_vram <= 4096:
        logging.warning(
            "Trying to enable lowvram mode because your GPU seems to have 4GB or less. If you don't want this use: --normalvram"
        )
        set_vram_to = VRAMState.LOW_VRAM

try:
    OOM_EXCEPTION = torch.cuda.OutOfMemoryError
except:
    OOM_EXCEPTION = Exception

XFORMERS_VERSION = ""
XFORMERS_ENABLED_VAE = True
if args.disable_xformers:
    XFORMERS_IS_AVAILABLE = False
else:
    try:
        import xformers
        import xformers.ops

        XFORMERS_IS_AVAILABLE = True
        try:
            XFORMERS_IS_AVAILABLE = xformers._has_cpp_library
        except:
            pass
        try:
            XFORMERS_VERSION = xformers.version.__version__
            logging.info("xformers version: {}".format(XFORMERS_VERSION))
            if XFORMERS_VERSION.startswith("0.0.18"):
                logging.warning(
                    "\nWARNING: This version of xformers has a major bug where you will get black images when generating high resolution images."
                )
                logging.warning(
                    "Please downgrade or upgrade xformers to a different version.\n"
                )
                XFORMERS_ENABLED_VAE = False
        except:
            pass
    except:
        XFORMERS_IS_AVAILABLE = False


def is_nvidia():
    global cpu_state
    if cpu_state == CPUState.GPU:
        if torch.version.cuda:
            return True
    return False


ENABLE_PYTORCH_ATTENTION = False
if args.use_pytorch_cross_attention:
    ENABLE_PYTORCH_ATTENTION = True
    XFORMERS_IS_AVAILABLE = False

VAE_DTYPE = torch.float32

try:
    if is_nvidia():
        torch_version = torch.version.__version__
        if int(torch_version[0]) >= 2:
            if (
                ENABLE_PYTORCH_ATTENTION == False
                and args.use_split_cross_attention == False
                and args.use_quad_cross_attention == False
            ):
                ENABLE_PYTORCH_ATTENTION = True
            if (
                torch.cuda.is_bf16_supported()
                and torch.cuda.get_device_properties(torch.cuda.current_device()).major
                >= 8
            ):
                VAE_DTYPE = torch.bfloat16
    if is_intel_xpu():
        if (
            args.use_split_cross_attention == False
            and args.use_quad_cross_attention == False
        ):
            ENABLE_PYTORCH_ATTENTION = True
except:
    pass

if is_intel_xpu():
    VAE_DTYPE = torch.bfloat16

if args.cpu_vae:
    VAE_DTYPE = torch.float32

if args.fp16_vae:
    VAE_DTYPE = torch.float16
elif args.bf16_vae:
    VAE_DTYPE = torch.bfloat16
elif args.fp32_vae:
    VAE_DTYPE = torch.float32


if ENABLE_PYTORCH_ATTENTION:
    torch.backends.cuda.enable_math_sdp(True)
    torch.backends.cuda.enable_flash_sdp(True)
    torch.backends.cuda.enable_mem_efficient_sdp(True)

if args.lowvram:
    set_vram_to = VRAMState.LOW_VRAM
    lowvram_available = True
elif args.novram:
    set_vram_to = VRAMState.NO_VRAM
elif args.highvram or args.gpu_only:
    vram_state = VRAMState.HIGH_VRAM

FORCE_FP32 = False
FORCE_FP16 = False
if args.force_fp32:
    logging.info("Forcing FP32, if this improves things please report it.")
    FORCE_FP32 = True

if args.force_fp16:
    logging.info("Forcing FP16.")
    FORCE_FP16 = True

if lowvram_available:
    if set_vram_to in (VRAMState.LOW_VRAM, VRAMState.NO_VRAM):
        vram_state = set_vram_to


if cpu_state != CPUState.GPU:
    vram_state = VRAMState.DISABLED

if cpu_state == CPUState.MPS:
    vram_state = VRAMState.SHARED

logging.info(f"Set vram state to: {vram_state.name}")

DISABLE_SMART_MEMORY = args.disable_smart_memory

if DISABLE_SMART_MEMORY:
    logging.info("Disabling smart memory management")


def get_torch_device_name(device):
    if hasattr(device, "type"):
        if device.type == "cuda":
            try:
                allocator_backend = torch.cuda.get_allocator_backend()
            except:
                allocator_backend = ""
            return "{} {} : {}".format(
                device, torch.cuda.get_device_name(device), allocator_backend
            )
        else:
            return "{}".format(device.type)
    elif is_intel_xpu():
        return "{} {}".format(device, torch.xpu.get_device_name(device))
    else:
        return "CUDA {}: {}".format(device, torch.cuda.get_device_name(device))


try:
    logging.info("Device: {}".format(get_torch_device_name(get_torch_device())))
except:
    logging.warning("Could not pick default device.")

logging.info("VAE dtype: {}".format(VAE_DTYPE))

current_loaded_models = []


def module_size(module):
    module_mem = 0
    sd = module.state_dict()
    for k in sd:
        t = sd[k]
        module_mem += t.nelement() * t.element_size()
    return module_mem


class LoadedModel:
    def __init__(self, model):
        self.model = model
        self.device = model.load_device
        self.weights_loaded = False
        self.real_model = None

    def model_memory(self):
        return self.model.model_size()

    def model_memory_required(self, device):
        if device == self.model.current_device:
            return 0
        else:
            return self.model_memory()

    def model_load(self, lowvram_model_memory=0):
        patch_model_to = self.device

        self.model.model_patches_to(self.device)
        self.model.model_patches_to(self.model.model_dtype())

        load_weights = not self.weights_loaded

        try:
            if lowvram_model_memory > 0 and load_weights:
                self.real_model = self.model.patch_model_lowvram(
                    device_to=patch_model_to, lowvram_model_memory=lowvram_model_memory
                )
            else:
                self.real_model = self.model.patch_model(
                    device_to=patch_model_to, patch_weights=load_weights
                )
        except Exception as e:
            self.model.unpatch_model(self.model.offload_device)
            self.model_unload()
            raise e

        if is_intel_xpu() and not args.disable_ipex_optimize:
            self.real_model = torch.xpu.optimize(
                self.real_model.eval(),
                inplace=True,
                auto_kernel_selection=True,
                graph_mode=True,
            )

        self.weights_loaded = True
        return self.real_model

    def model_unload(self, unpatch_weights=True):
        self.model.unpatch_model(
            self.model.offload_device, unpatch_weights=unpatch_weights
        )
        self.model.model_patches_to(self.model.offload_device)
        self.weights_loaded = self.weights_loaded and not unpatch_weights
        self.real_model = None

    def __eq__(self, other):
        return self.model is other.model


def minimum_inference_memory():
    return 1024 * 1024 * 1024


def unload_model_clones(model, unload_weights_only=True, force_unload=True):
    to_unload = []
    for i in range(len(current_loaded_models)):
        if model.is_clone(current_loaded_models[i].model):
            to_unload = [i] + to_unload

    if len(to_unload) == 0:
        return True

    same_weights = 0
    for i in to_unload:
        if model.clone_has_same_weights(current_loaded_models[i].model):
            same_weights += 1

    if same_weights == len(to_unload):
        unload_weight = False
    else:
        unload_weight = True

    if not force_unload:
        if unload_weights_only and unload_weight == False:
            return None

    for i in to_unload:
        logging.debug("unload clone {} {}".format(i, unload_weight))
        current_loaded_models.pop(i).model_unload(unpatch_weights=unload_weight)

    return unload_weight


def free_memory(memory_required, device, keep_loaded=[]):
    unloaded_model = []
    can_unload = []

    for i in range(len(current_loaded_models) - 1, -1, -1):
        shift_model = current_loaded_models[i]
        if shift_model.device == device:
            if shift_model not in keep_loaded:
                can_unload.append(
                    (sys.getrefcount(shift_model.model), shift_model.model_memory(), i)
                )

    for x in sorted(can_unload):
        i = x[-1]
        if not DISABLE_SMART_MEMORY:
            if get_free_memory(device) > memory_required:
                break
        current_loaded_models[i].model_unload()
        unloaded_model.append(i)

    for i in sorted(unloaded_model, reverse=True):
        current_loaded_models.pop(i)

    if len(unloaded_model) > 0:
        soft_empty_cache()
    else:
        if vram_state != VRAMState.HIGH_VRAM:
            mem_free_total, mem_free_torch = get_free_memory(
                device, torch_free_too=True
            )
            if mem_free_torch > mem_free_total * 0.25:
                soft_empty_cache()


def load_models_gpu(models, memory_required=0):
    global vram_state

    inference_memory = minimum_inference_memory()
    extra_mem = max(inference_memory, memory_required)

    models_to_load = []
    models_already_loaded = []
    for x in models:
        loaded_model = LoadedModel(x)

        if loaded_model in current_loaded_models:
            index = current_loaded_models.index(loaded_model)
            current_loaded_models.insert(0, current_loaded_models.pop(index))
            models_already_loaded.append(loaded_model)
        else:
            if hasattr(x, "model"):
                logging.info(f"Requested to load {x.model.__class__.__name__}")
            models_to_load.append(loaded_model)

    if len(models_to_load) == 0:
        devs = set(map(lambda a: a.device, models_already_loaded))
        for d in devs:
            if d != torch.device("cpu"):
                free_memory(extra_mem, d, models_already_loaded)
        return

    logging.info(
        f"Loading {len(models_to_load)} new model{'s' if len(models_to_load) > 1 else ''}"
    )

    total_memory_required = {}
    for loaded_model in models_to_load:
        if (
            unload_model_clones(
                loaded_model.model, unload_weights_only=True, force_unload=False
            )
            == True
        ):  # unload clones where the weights are different
            total_memory_required[loaded_model.device] = total_memory_required.get(
                loaded_model.device, 0
            ) + loaded_model.model_memory_required(loaded_model.device)

    for device in total_memory_required:
        if device != torch.device("cpu"):
            free_memory(
                total_memory_required[device] * 1.3 + extra_mem,
                device,
                models_already_loaded,
            )

    for loaded_model in models_to_load:
        weights_unloaded = unload_model_clones(
            loaded_model.model, unload_weights_only=False, force_unload=False
        )  # unload the rest of the clones where the weights can stay loaded
        if weights_unloaded is not None:
            loaded_model.weights_loaded = not weights_unloaded

    for loaded_model in models_to_load:
        model = loaded_model.model
        torch_dev = model.load_device
        if is_device_cpu(torch_dev):
            vram_set_state = VRAMState.DISABLED
        else:
            vram_set_state = vram_state
        lowvram_model_memory = 0
        if lowvram_available and (
            vram_set_state == VRAMState.LOW_VRAM
            or vram_set_state == VRAMState.NORMAL_VRAM
        ):
            model_size = loaded_model.model_memory_required(torch_dev)
            current_free_mem = get_free_memory(torch_dev)
            lowvram_model_memory = int(
                max(64 * (1024 * 1024), (current_free_mem - 1024 * (1024 * 1024)) / 1.3)
            )
            if model_size > (
                current_free_mem - inference_memory
            ):  # only switch to lowvram if really necessary
                vram_set_state = VRAMState.LOW_VRAM
            else:
                lowvram_model_memory = 0

        if vram_set_state == VRAMState.NO_VRAM:
            lowvram_model_memory = 64 * 1024 * 1024

        cur_loaded_model = loaded_model.model_load(lowvram_model_memory)
        current_loaded_models.insert(0, loaded_model)
    return


def load_model_gpu(model):
    return load_models_gpu([model])


def cleanup_models(keep_clone_weights_loaded=False):
    to_delete = []
    for i in range(len(current_loaded_models)):
        if sys.getrefcount(current_loaded_models[i].model) <= 2:
            if not keep_clone_weights_loaded:
                to_delete = [i] + to_delete
            # TODO: find a less fragile way to do this.
            elif (
                sys.getrefcount(current_loaded_models[i].real_model) <= 3
            ):  # references from .real_model + the .model
                to_delete = [i] + to_delete

    for i in to_delete:
        x = current_loaded_models.pop(i)
        x.model_unload()
        del x


def dtype_size(dtype):
    dtype_size = 4
    if dtype == torch.float16 or dtype == torch.bfloat16:
        dtype_size = 2
    elif dtype == torch.float32:
        dtype_size = 4
    else:
        try:
            dtype_size = dtype.itemsize
        except:  # Old pytorch doesn't have .itemsize
            pass
    return dtype_size


def unet_offload_device():
    if vram_state == VRAMState.HIGH_VRAM:
        return get_torch_device()
    else:
        return torch.device("cpu")


def unet_inital_load_device(parameters, dtype):
    torch_dev = get_torch_device()
    if vram_state == VRAMState.HIGH_VRAM:
        return torch_dev

    cpu_dev = torch.device("cpu")
    if DISABLE_SMART_MEMORY:
        return cpu_dev

    model_size = dtype_size(dtype) * parameters

    mem_dev = get_free_memory(torch_dev)
    mem_cpu = get_free_memory(cpu_dev)
    if mem_dev > mem_cpu and model_size < mem_dev:
        return torch_dev
    else:
        return cpu_dev


def unet_dtype(
    device=None,
    model_params=0,
    supported_dtypes=[torch.float16, torch.bfloat16, torch.float32],
):
    if args.bf16_unet:
        return torch.bfloat16
    if args.fp16_unet:
        return torch.float16
    if args.fp8_e4m3fn_unet:
        return torch.float8_e4m3fn
    if args.fp8_e5m2_unet:
        return torch.float8_e5m2
    if should_use_fp16(device=device, model_params=model_params, manual_cast=True):
        if torch.float16 in supported_dtypes:
            return torch.float16
    if should_use_bf16(device, model_params=model_params, manual_cast=True):
        if torch.bfloat16 in supported_dtypes:
            return torch.bfloat16
    return torch.float32


# None means no manual cast
def unet_manual_cast(
    weight_dtype,
    inference_device,
    supported_dtypes=[torch.float16, torch.bfloat16, torch.float32],
):
    if weight_dtype == torch.float32:
        return None

    fp16_supported = should_use_fp16(inference_device, prioritize_performance=False)
    if fp16_supported and weight_dtype == torch.float16:
        return None

    bf16_supported = should_use_bf16(inference_device)
    if bf16_supported and weight_dtype == torch.bfloat16:
        return None

    if fp16_supported and torch.float16 in supported_dtypes:
        return torch.float16

    elif bf16_supported and torch.bfloat16 in supported_dtypes:
        return torch.bfloat16
    else:
        return torch.float32


def text_encoder_offload_device():
    if args.gpu_only:
        return get_torch_device()
    else:
        return torch.device("cpu")


def text_encoder_device():
    if args.gpu_only:
        return get_torch_device()
    elif vram_state == VRAMState.HIGH_VRAM or vram_state == VRAMState.NORMAL_VRAM:
        if is_intel_xpu():
            return torch.device("cpu")
        if should_use_fp16(prioritize_performance=False):
            return get_torch_device()
        else:
            return torch.device("cpu")
    else:
        return torch.device("cpu")


def text_encoder_dtype(device=None):
    if args.fp8_e4m3fn_text_enc:
        return torch.float8_e4m3fn
    elif args.fp8_e5m2_text_enc:
        return torch.float8_e5m2
    elif args.fp16_text_enc:
        return torch.float16
    elif args.fp32_text_enc:
        return torch.float32

    if is_device_cpu(device):
        return torch.float16

    return torch.float16


def intermediate_device():
    if args.gpu_only:
        return get_torch_device()
    else:
        return torch.device("cpu")


def vae_device():
    if args.cpu_vae:
        return torch.device("cpu")
    return get_torch_device()


def vae_offload_device():
    if args.gpu_only:
        return get_torch_device()
    else:
        return torch.device("cpu")


def vae_dtype():
    global VAE_DTYPE
    return VAE_DTYPE


def get_autocast_device(dev):
    if hasattr(dev, "type"):
        return dev.type
    return "cuda"


def supports_dtype(device, dtype):  # TODO
    if dtype == torch.float32:
        return True
    if is_device_cpu(device):
        return False
    if dtype == torch.float16:
        return True
    if dtype == torch.bfloat16:
        return True
    return False


def device_supports_non_blocking(device):
    if is_device_mps(device):
        return False  # pytorch bug? mps doesn't support non blocking
    return True


def cast_to_device(tensor, device, dtype, copy=False):
    device_supports_cast = False
    if tensor.dtype == torch.float32 or tensor.dtype == torch.float16:
        device_supports_cast = True
    elif tensor.dtype == torch.bfloat16:
        if hasattr(device, "type") and device.type.startswith("cuda"):
            device_supports_cast = True
        elif is_intel_xpu():
            device_supports_cast = True

    non_blocking = device_supports_non_blocking(device)

    if device_supports_cast:
        if copy:
            if tensor.device == device:
                return tensor.to(dtype, copy=copy, non_blocking=non_blocking)
            return tensor.to(device, copy=copy, non_blocking=non_blocking).to(
                dtype, non_blocking=non_blocking
            )
        else:
            return tensor.to(device, non_blocking=non_blocking).to(
                dtype, non_blocking=non_blocking
            )
    else:
        return tensor.to(device, dtype, copy=copy, non_blocking=non_blocking)


def xformers_enabled():
    global directml_enabled
    global cpu_state
    if cpu_state != CPUState.GPU:
        return False
    if is_intel_xpu():
        return False
    if directml_enabled:
        return False
    return XFORMERS_IS_AVAILABLE


def xformers_enabled_vae():
    enabled = xformers_enabled()
    if not enabled:
        return False

    return XFORMERS_ENABLED_VAE


def pytorch_attention_enabled():
    global ENABLE_PYTORCH_ATTENTION
    return ENABLE_PYTORCH_ATTENTION


def pytorch_attention_flash_attention():
    global ENABLE_PYTORCH_ATTENTION
    if ENABLE_PYTORCH_ATTENTION:
        # TODO: more reliable way of checking for flash attention?
        if is_nvidia():  # pytorch flash attention only works on Nvidia
            return True
    return False


def get_free_memory(dev=None, torch_free_too=False):
    global directml_enabled
    if dev is None:
        dev = get_torch_device()

    if hasattr(dev, "type") and (dev.type == "cpu" or dev.type == "mps"):
        mem_free_total = psutil.virtual_memory().available
        mem_free_torch = mem_free_total
    else:
        if directml_enabled:
            mem_free_total = 1024 * 1024 * 1024  # TODO
            mem_free_torch = mem_free_total
        elif is_intel_xpu():
            stats = torch.xpu.memory_stats(dev)
            mem_active = stats["active_bytes.all.current"]
            mem_allocated = stats["allocated_bytes.all.current"]
            mem_reserved = stats["reserved_bytes.all.current"]
            mem_free_torch = mem_reserved - mem_active
            mem_free_total = (
                torch.xpu.get_device_properties(dev).total_memory - mem_allocated
            )
        else:
            stats = torch.cuda.memory_stats(dev)
            mem_active = stats["active_bytes.all.current"]
            mem_reserved = stats["reserved_bytes.all.current"]
            mem_free_cuda, _ = torch.cuda.mem_get_info(dev)
            mem_free_torch = mem_reserved - mem_active
            mem_free_total = mem_free_cuda + mem_free_torch

    if torch_free_too:
        return (mem_free_total, mem_free_torch)
    else:
        return mem_free_total


def cpu_mode():
    global cpu_state
    return cpu_state == CPUState.CPU


def mps_mode():
    global cpu_state
    return cpu_state == CPUState.MPS


def is_device_type(device, type):
    if hasattr(device, "type"):
        if device.type == type:
            return True
    return False


def is_device_cpu(device):
    return is_device_type(device, "cpu")


def is_device_mps(device):
    return is_device_type(device, "mps")


def is_device_cuda(device):
    return is_device_type(device, "cuda")


def should_use_fp16(
    device=None, model_params=0, prioritize_performance=True, manual_cast=False
):
    global directml_enabled

    if device is not None:
        if is_device_cpu(device):
            return False

    if FORCE_FP16:
        return True

    if device is not None:
        if is_device_mps(device):
            return True

    if FORCE_FP32:
        return False

    if directml_enabled:
        return False

    if mps_mode():
        return True

    if cpu_mode():
        return False

    if is_intel_xpu():
        return True

    if torch.version.hip:
        return True

    props = torch.cuda.get_device_properties("cuda")
    if props.major >= 8:
        return True

    if props.major < 6:
        return False

    fp16_works = False
    # FP16 is confirmed working on a 1080 (GP104) but it's a bit slower than FP32 so it should only be enabled
    # when the model doesn't actually fit on the card
    # TODO: actually test if GP106 and others have the same type of behavior
    nvidia_10_series = [
        "1080",
        "1070",
        "titan x",
        "p3000",
        "p3200",
        "p4000",
        "p4200",
        "p5000",
        "p5200",
        "p6000",
        "1060",
        "1050",
        "p40",
        "p100",
        "p6",
        "p4",
    ]
    for x in nvidia_10_series:
        if x in props.name.lower():
            fp16_works = True

    if fp16_works or manual_cast:
        free_model_memory = get_free_memory() * 0.9 - minimum_inference_memory()
        if (not prioritize_performance) or model_params * 4 > free_model_memory:
            return True

    if props.major < 7:
        return False

    # FP16 is just broken on these cards
    nvidia_16_series = [
        "1660",
        "1650",
        "1630",
        "T500",
        "T550",
        "T600",
        "MX550",
        "MX450",
        "CMP 30HX",
        "T2000",
        "T1000",
        "T1200",
    ]
    for x in nvidia_16_series:
        if x in props.name:
            return False

    return True


def should_use_bf16(
    device=None, model_params=0, prioritize_performance=True, manual_cast=False
):
    if device is not None:
        if is_device_cpu(device):  # TODO ? bf16 works on CPU but is extremely slow
            return False

    if device is not None:  # TODO not sure about mps bf16 support
        if is_device_mps(device):
            return False

    if FORCE_FP32:
        return False

    if directml_enabled:
        return False

    if cpu_mode() or mps_mode():
        return False

    if is_intel_xpu():
        return True

    if device is None:
        device = torch.device("cuda")

    props = torch.cuda.get_device_properties(device)
    if props.major >= 8:
        return True

    bf16_works = torch.cuda.is_bf16_supported()

    if bf16_works or manual_cast:
        free_model_memory = get_free_memory() * 0.9 - minimum_inference_memory()
        if (not prioritize_performance) or model_params * 4 > free_model_memory:
            return True

    return False


def soft_empty_cache(force=False):
    global cpu_state
    if cpu_state == CPUState.MPS:
        torch.mps.empty_cache()
    elif is_intel_xpu():
        torch.xpu.empty_cache()
    elif torch.cuda.is_available():
        if (
            force or is_nvidia()
        ):  # This seems to make things worse on ROCm so I only do it for cuda
            torch.cuda.empty_cache()
            torch.cuda.ipc_collect()


def unload_all_models():
    free_memory(1e30, get_torch_device())


def resolve_lowvram_weight(weight, model, key):  # TODO: remove
    return weight


# TODO: might be cleaner to put this somewhere else
import threading


class InterruptProcessingException(Exception):
    pass


interrupt_processing_mutex = threading.RLock()

interrupt_processing = False


def interrupt_current_processing(value=True):
    global interrupt_processing
    global interrupt_processing_mutex
    with interrupt_processing_mutex:
        interrupt_processing = value


def processing_interrupted():
    global interrupt_processing
    global interrupt_processing_mutex
    with interrupt_processing_mutex:
        return interrupt_processing


def throw_exception_if_processing_interrupted():
    global interrupt_processing
    global interrupt_processing_mutex
    with interrupt_processing_mutex:
        if interrupt_processing:
            interrupt_processing = False
            raise InterruptProcessingException()



import os

from nodetool.workflows.workflow_node import WorkflowNode


current_dir = os.path.dirname(os.path.realpath(__file__))
workflows_dir = os.path.join(current_dir, "workflows")


class StableDiffusion(WorkflowNode):
    @classmethod
    def get_workflow_file(cls) -> str:
        return os.path.join(workflows_dir, "stable_diffusion.json")



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode

from nodetool.nodes.huggingface.image.generate import ModelId
from nodetool.nodes.huggingface.image.generate import Scheduler

class StableDiffusionXL(GraphNode):
    model: ModelId | GraphNode | tuple[GraphNode, str] = Field(default=ModelId('stabilityai/stable-diffusion-xl-base-1.0'), description='The model ID to use for the image generation')
    inputs: str | GraphNode | tuple[GraphNode, str] = Field(default='A photo of a cat.', description='The input text to the model')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The negative prompt to use.')
    seed: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description=None)
    guidance_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=7.0, description=None)
    num_inference_steps: int | GraphNode | tuple[GraphNode, str] = Field(default=30, description=None)
    width: int | GraphNode | tuple[GraphNode, str] = Field(default=768, description=None)
    height: int | GraphNode | tuple[GraphNode, str] = Field(default=768, description=None)
    scheduler: Scheduler | GraphNode | tuple[GraphNode, str] = Field(default=Scheduler('EulerDiscreteScheduler'), description=None)
    @classmethod
    def get_node_type(cls): return "huggingface.image.generate.StableDiffusionXL"





import os
from .api import make_detectron2_model, semantic_run
from pathlib import Path
import warnings
from controlnet_aux.util import HWC3, common_input_validate, resize_image_with_pad, custom_hf_download, HF_MODEL_NAME
import numpy as np
import cv2
from PIL import Image

DEFAULT_CONFIGS = {
    "coco": {
        "name": "150_16_swin_l_oneformer_coco_100ep.pth",
        "config": os.path.join(os.path.dirname(__file__), 'configs/coco/oneformer_swin_large_IN21k_384_bs16_100ep.yaml')
    },
    "ade20k": {
        "name": "250_16_swin_l_oneformer_ade20k_160k.pth",
        "config": os.path.join(os.path.dirname(__file__), 'configs/ade20k/oneformer_swin_large_IN21k_384_bs16_160k.yaml')
    }
}
class OneformerSegmentor:
    def __init__(self, model, metadata):
        self.model = model
        self.metadata = metadata

    def to(self, device):
        self.model.model.to(device)
        return self
    
    @classmethod
    def from_pretrained(cls, pretrained_model_or_path=HF_MODEL_NAME, filename="250_16_swin_l_oneformer_ade20k_160k.pth", config_path = None):
        config_path = config_path or DEFAULT_CONFIGS["ade20k" if "ade20k" in filename else "coco"]["config"]
        model_path = custom_hf_download(pretrained_model_or_path, filename)

        model, metadata = make_detectron2_model(config_path, model_path)

        return cls(model, metadata)
    
    def __call__(self, input_image=None, detect_resolution=512, output_type=None, upscale_method="INTER_CUBIC", **kwargs):
        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        input_image, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)

        detected_map = semantic_run(input_image, self.model, self.metadata)
        detected_map = remove_pad(HWC3(detected_map))
        
        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)
            
        return detected_map



from ..utils import common_annotator_call, create_node_input_types
import comfy.model_management as model_management

class OpenPose_Preprocessor:
    @classmethod
    def INPUT_TYPES(s):
        return create_node_input_types(
            detect_hand = (["enable", "disable"], {"default": "enable"}),
            detect_body = (["enable", "disable"], {"default": "enable"}),
            detect_face = (["enable", "disable"], {"default": "enable"})
        )
        
    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "estimate_pose"

    CATEGORY = "ControlNet Preprocessors/Faces and Poses Estimators"

    def estimate_pose(self, image, detect_hand, detect_body, detect_face, resolution=512, **kwargs):
        from controlnet_aux.open_pose import OpenposeDetector

        detect_hand = detect_hand == "enable"
        detect_body = detect_body == "enable"
        detect_face = detect_face == "enable"


        self.openpose_json = None
        model = OpenposeDetector.from_pretrained().to(model_management.get_torch_device())
        
        def cb(image, **kwargs):
            result = model(image, **kwargs)
            self.openpose_json = result[1]
            return result[0]
        
        out = common_annotator_call(cb, image, include_hand=detect_hand, include_face=detect_face, include_body=detect_body, image_and_json=True, resolution=resolution)
        del model
        return {
            'ui': { "openpose_json": [self.openpose_json] },
            "result": (out, )
        }

NODE_CLASS_MAPPINGS = {
    "OpenposePreprocessor": OpenPose_Preprocessor,
}
NODE_DISPLAY_NAME_MAPPINGS = {
    "OpenposePreprocessor": "OpenPose Pose",
}


import cv2
import numpy as np
from PIL import Image
from controlnet_aux.util import resize_image_with_pad, common_input_validate, HWC3

class LineartStandardDetector:
    def __call__(self, input_image=None, guassian_sigma=6.0, intensity_threshold=8, detect_resolution=512, output_type=None, upscale_method="INTER_CUBIC", **kwargs):
        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        input_image, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)
        
        x = input_image.astype(np.float32)
        g = cv2.GaussianBlur(x, (0, 0), guassian_sigma)
        intensity = np.min(g - x, axis=2).clip(0, 255)
        intensity /= max(16, np.median(intensity[intensity > intensity_threshold]))
        intensity *= 127
        detected_map = intensity.clip(0, 255).astype(np.uint8)
        
        detected_map = HWC3(remove_pad(detected_map))
        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)
        return detected_map


from enum import Enum
from pydantic import Field
from nodetool.metadata.types import RankingResult
from nodetool.workflows.base_node import BaseNode
from nodetool.workflows.processing_context import ProcessingContext


class Model(str, Enum):
    MXBAI = "mixedbread-ai/mxbai-rerank-large-v1"


class Rerank(BaseNode):
    model: Model = Field(default=Model.MXBAI, description="The reranking model to use")
    query: str = Field(default="", description="The question to rerank")
    documents: list[str] = Field(
        default=[], description="The answers to be ranked by the model."
    )
    top_k: int

    async def process(self, context: ProcessingContext) -> list[RankingResult]:
        from sentence_transformers import CrossEncoder

        model = CrossEncoder("mixedbread-ai/mxbai-rerank-large-v1")
        results = model.rank(
            self.query, self.documents, return_documents=True, top_k=self.top_k
        )
        return [RankingResult(score=res["score"], text=res["text"]) for res in results]



from enum import Enum
from typing import Any
from pydantic import Field
from nodetool.metadata.types import ImageTensor
from nodetool.common.comfy_node import ComfyNode


class SaveAnimatedWEBP(ComfyNode):
    images: list[ImageTensor] = Field(
        default_factory=list, description="list of images to save as animated WEBP."
    )
    filename_prefix: str = Field(
        default="ComfyUI", description="Prefix for the filename."
    )
    fps: float = Field(default=6.0, description="Frames per second.")
    lossless: bool = Field(
        default=True, description="Whether to use lossless compression."
    )
    quality: int = Field(default=80, description="Quality of the WEBP.")
    method: Enum = Field(
        default="default", description="Compression method to use."
    )  # Enum should be adjusted to have the correct values from the "methods" list.

    hidden_fields: dict[str, Any] = Field(
        default_factory=dict,
        description="Hidden fields like prompt and extra PNG information.",
    )
    output_node: bool = True

    @classmethod
    def return_type(cls):
        return None


class SaveAnimatedPNG(ComfyNode):
    images: list[ImageTensor] = Field(
        default_factory=list, description="list of images to save as animated PNG."
    )
    filename_prefix: str = Field(
        default="ComfyUI", description="Prefix for the filename."
    )
    fps: float = Field(default=6.0, description="Frames per second.")
    compress_level: int = Field(default=4, description="PNG compression level.")

    hidden_fields: dict[str, Any] = Field(
        default_factory=dict,
        description="Hidden fields like prompt and extra PNG information.",
    )
    output_node: bool = True

    @classmethod
    def return_type(cls):
        return None



from datetime import datetime
from nodetool.metadata.types import DataframeRef
from nodetool.nodes.nodetool.input import GroupInput
from nodetool.nodes.nodetool.output import GroupOutput
from nodetool.workflows.base_node import GroupNode
from nodetool.workflows.graph import Graph
from nodetool.workflows.processing_context import ProcessingContext

from typing import Any

from nodetool.workflows.types import NodeUpdate
from nodetool.workflows.workflow_runner import WorkflowRunner


class Loop(GroupNode):
    """
    Loops over a list of items and processes the remaining nodes for each item.
    loop, itereate, repeat, for, each, batch
    """

    async def process(self, context: ProcessingContext):
        raise NotImplementedError()

    async def process_subgraph(
        self,
        context: ProcessingContext,
        runner: WorkflowRunner,
    ) -> Any:
        """
        Runs a loop node, which is a special type of node that contains a subgraph.
        The subgraph is determined by the children property of the group node.
        Each iteration of the loop will run the subgraph with the input data from the input nodes.
        The output will be collected and stored in the output nodes.
        """
        input_nodes = [n for n in self._nodes if isinstance(n, GroupInput)]
        output_nodes = [n for n in self._nodes if isinstance(n, GroupOutput)]

        for name, value in self._properties.items():
            if isinstance(value, DataframeRef):
                df = await context.dataframe_to_pandas(value)
                self._properties[name] = df.to_dict(orient="records")

        if len(input_nodes) == 0:
            raise ValueError("Loop node must have at least one input node.")

        # Validate all inputs to be list
        assert all(
            isinstance(value, list) for value in self._properties.values()
        ), "Input data must be a list or dataframe."

        # Get the length of the input data and check for equal length.
        lengths = [len(value) for value in self._properties.values()]
        if len(input_nodes) > 1 and len(set(lengths)) != 1:
            raise ValueError("Input data must be of equal length.")

        input_length = lengths[0]
        started_at = datetime.now()

        # Run the subgraph for each item in the input data.
        results = {node._id: [] for node in output_nodes}
        for i in range(input_length):
            sub_context = ProcessingContext(
                user_id=context.user_id,
                auth_token=context.auth_token,
                workflow_id=context.workflow_id,
                queue=context.message_queue,
                capabilities=context.capabilities,
            )
            # Provide the item as input to the subgraph.
            for input_node in input_nodes:
                if input_node.name in self._properties:
                    input_sequence = self._properties[input_node.name]
                    if len(input_sequence) > i:
                        input_node._value = input_sequence[i]

            graph = Graph(nodes=self._nodes, edges=self._edges)
            await runner.process_graph(sub_context, graph)

            # Get the result of the subgraph and add it to the results.
            for output_node in output_nodes:
                results[output_node._id].append(output_node.input)

        for output_node in output_nodes:
            context.set_result(self._id, {output_node.name: results[output_node._id]})
            context.post_message(
                NodeUpdate(
                    node_id=output_node._id,
                    node_name=output_node.get_title(),
                    status="completed",
                    result={"output": results[output_node._id]},
                    started_at=started_at.isoformat(),
                    completed_at=datetime.now().isoformat(),
                )
            )

        # Mark the nodes as processed.
        for n in self._nodes:
            context.processed_nodes.add(n._id)



"""
Copyright (c) Microsoft Corporation.
Licensed under the MIT license.

End-to-end inference codes for 
3D human body mesh reconstruction from an image
"""

from __future__ import absolute_import, division, print_function
import argparse
import os
import os.path as op
import code
import json
import time
import datetime
import torch
import torchvision.models as models
from torchvision.utils import make_grid
import gc
import numpy as np
import cv2
from custom_mesh_graphormer.modeling.bert import BertConfig, Graphormer
from custom_mesh_graphormer.modeling.bert import Graphormer_Body_Network as Graphormer_Network
from custom_mesh_graphormer.modeling._smpl import SMPL, Mesh
from custom_mesh_graphormer.modeling.hrnet.hrnet_cls_net_gridfeat import get_cls_net_gridfeat
from custom_mesh_graphormer.modeling.hrnet.config import config as hrnet_config
from custom_mesh_graphormer.modeling.hrnet.config import update_config as hrnet_update_config
import custom_mesh_graphormer.modeling.data.config as cfg
from custom_mesh_graphormer.datasets.build import make_data_loader

from custom_mesh_graphormer.utils.logger import setup_logger
from custom_mesh_graphormer.utils.comm import synchronize, is_main_process, get_rank, get_world_size, all_gather
from custom_mesh_graphormer.utils.miscellaneous import mkdir, set_seed
from custom_mesh_graphormer.utils.metric_logger import AverageMeter, EvalMetricsLogger
from custom_mesh_graphormer.utils.renderer import Renderer, visualize_reconstruction_and_att_local, visualize_reconstruction_no_text
from custom_mesh_graphormer.utils.metric_pampjpe import reconstruction_error
from custom_mesh_graphormer.utils.geometric_layers import orthographic_projection

from PIL import Image
from torchvision import transforms

from comfy.model_management import get_torch_device
device = get_torch_device()

transform = transforms.Compose([           
                    transforms.Resize(224),
                    transforms.CenterCrop(224),
                    transforms.ToTensor(),
                    transforms.Normalize(
                        mean=[0.485, 0.456, 0.406],
                        std=[0.229, 0.224, 0.225])])

transform_visualize = transforms.Compose([           
                    transforms.Resize(224),
                    transforms.CenterCrop(224),
                    transforms.ToTensor()])

def run_inference(args, image_list, Graphormer_model, smpl, renderer, mesh_sampler):
    # switch to evaluate mode
    Graphormer_model.eval()
    smpl.eval()
    with torch.no_grad():
        for image_file in image_list:
            if 'pred' not in image_file:
                att_all = []
                img = Image.open(image_file)
                img_tensor = transform(img)
                img_visual = transform_visualize(img)

                batch_imgs = torch.unsqueeze(img_tensor, 0).to(device)
                batch_visual_imgs = torch.unsqueeze(img_visual, 0).to(device)
                # forward-pass
                pred_camera, pred_3d_joints, pred_vertices_sub2, pred_vertices_sub, pred_vertices, hidden_states, att = Graphormer_model(batch_imgs, smpl, mesh_sampler)
                    
                # obtain 3d joints from full mesh
                pred_3d_joints_from_smpl = smpl.get_h36m_joints(pred_vertices)

                pred_3d_pelvis = pred_3d_joints_from_smpl[:,cfg.H36M_J17_NAME.index('Pelvis'),:]
                pred_3d_joints_from_smpl = pred_3d_joints_from_smpl[:,cfg.H36M_J17_TO_J14,:]
                pred_3d_joints_from_smpl = pred_3d_joints_from_smpl - pred_3d_pelvis[:, None, :]
                pred_vertices = pred_vertices - pred_3d_pelvis[:, None, :]

                # save attantion
                att_max_value = att[-1]
                att_cpu = np.asarray(att_max_value.cpu().detach())
                att_all.append(att_cpu)

                # obtain 3d joints, which are regressed from the full mesh
                pred_3d_joints_from_smpl = smpl.get_h36m_joints(pred_vertices)
                pred_3d_joints_from_smpl = pred_3d_joints_from_smpl[:,cfg.H36M_J17_TO_J14,:]
                # obtain 2d joints, which are projected from 3d joints of smpl mesh
                pred_2d_joints_from_smpl = orthographic_projection(pred_3d_joints_from_smpl, pred_camera)
                pred_2d_431_vertices_from_smpl = orthographic_projection(pred_vertices_sub2, pred_camera)
                visual_imgs_output = visualize_mesh( renderer, batch_visual_imgs[0],
                                                                pred_vertices[0].detach(), 
                                                                pred_camera.detach())
                # visual_imgs_output = visualize_mesh_and_attention( renderer, batch_visual_imgs[0],
                #                                             pred_vertices[0].detach(), 
                #                                             pred_vertices_sub2[0].detach(), 
                #                                             pred_2d_431_vertices_from_smpl[0].detach(),
                #                                             pred_2d_joints_from_smpl[0].detach(),
                #                                             pred_camera.detach(),
                #                                             att[-1][0].detach())

                visual_imgs = visual_imgs_output.transpose(1,2,0)
                visual_imgs = np.asarray(visual_imgs)
                        
                temp_fname = image_file[:-4] + '_graphormer_pred.jpg'
                print('save to ', temp_fname)
                cv2.imwrite(temp_fname, np.asarray(visual_imgs[:,:,::-1]*255))

    return 

def visualize_mesh( renderer, images,
                    pred_vertices_full,
                    pred_camera):
    img = images.cpu().numpy().transpose(1,2,0)
    # Get predict vertices for the particular example
    vertices_full = pred_vertices_full.cpu().numpy() 
    cam = pred_camera.cpu().numpy()
    # Visualize only mesh reconstruction 
    rend_img = visualize_reconstruction_no_text(img, 224, vertices_full, cam, renderer, color='light_blue')
    rend_img = rend_img.transpose(2,0,1)
    return rend_img

def visualize_mesh_and_attention( renderer, images,
                    pred_vertices_full,
                    pred_vertices, 
                    pred_2d_vertices,
                    pred_2d_joints,
                    pred_camera,
                    attention):
    img = images.cpu().numpy().transpose(1,2,0)
    # Get predict vertices for the particular example
    vertices_full = pred_vertices_full.cpu().numpy() 
    vertices = pred_vertices.cpu().numpy()
    vertices_2d = pred_2d_vertices.cpu().numpy()
    joints_2d = pred_2d_joints.cpu().numpy()
    cam = pred_camera.cpu().numpy()
    att = attention.cpu().numpy()
    # Visualize reconstruction and attention
    rend_img = visualize_reconstruction_and_att_local(img, 224, vertices_full, vertices, vertices_2d, cam, renderer, joints_2d, att, color='light_blue')
    rend_img = rend_img.transpose(2,0,1)
    return rend_img


def parse_args():
    parser = argparse.ArgumentParser()
    #########################################################
    # Data related arguments
    #########################################################
    parser.add_argument("--num_workers", default=4, type=int, 
                        help="Workers in dataloader.")
    parser.add_argument("--img_scale_factor", default=1, type=int, 
                        help="adjust image resolution.")
    parser.add_argument("--image_file_or_path", default='./samples/human-body', type=str, 
                        help="test data") 
    #########################################################
    # Loading/saving checkpoints
    #########################################################
    parser.add_argument("--model_name_or_path", default='src/modeling/bert/bert-base-uncased/', type=str, required=False,
                        help="Path to pre-trained transformer model or model type.")
    parser.add_argument("--resume_checkpoint", default=None, type=str, required=False,
                        help="Path to specific checkpoint for resume training.")
    parser.add_argument("--output_dir", default='output/', type=str, required=False,
                        help="The output directory to save checkpoint and test results.")
    parser.add_argument("--config_name", default="", type=str, 
                        help="Pretrained config name or path if not the same as model_name.")
    #########################################################
    # Model architectures
    #########################################################
    parser.add_argument('-a', '--arch', default='hrnet-w64',
                    help='CNN backbone architecture: hrnet-w64, hrnet, resnet50')
    parser.add_argument("--num_hidden_layers", default=4, type=int, required=False, 
                        help="Update model config if given")
    parser.add_argument("--hidden_size", default=-1, type=int, required=False, 
                        help="Update model config if given")
    parser.add_argument("--num_attention_heads", default=4, type=int, required=False, 
                        help="Update model config if given. Note that the division of "
                        "hidden_size / num_attention_heads should be in integer.")
    parser.add_argument("--intermediate_size", default=-1, type=int, required=False, 
                        help="Update model config if given.")
    parser.add_argument("--input_feat_dim", default='2051,512,128', type=str, 
                        help="The Image Feature Dimension.")          
    parser.add_argument("--hidden_feat_dim", default='1024,256,64', type=str, 
                        help="The Image Feature Dimension.")   
    parser.add_argument("--which_gcn", default='0,0,1', type=str, 
                        help="which encoder block to have graph conv. Encoder1, Encoder2, Encoder3. Default: only Encoder3 has graph conv") 
    parser.add_argument("--mesh_type", default='body', type=str, help="body or hand") 
    parser.add_argument("--interm_size_scale", default=2, type=int)
    #########################################################
    # Others
    #########################################################
    parser.add_argument("--run_eval_only", default=True, action='store_true',) 
    parser.add_argument("--device", type=str, default='cuda', 
                        help="cuda or cpu")
    parser.add_argument('--seed', type=int, default=88, 
                        help="random seed for initialization.")

    args = parser.parse_args()
    return args


def main(args):
    global logger
    # Setup CUDA, GPU & distributed training
    args.num_gpus = int(os.environ['WORLD_SIZE']) if 'WORLD_SIZE' in os.environ else 1
    os.environ['OMP_NUM_THREADS'] = str(args.num_workers)
    print('set os.environ[OMP_NUM_THREADS] to {}'.format(os.environ['OMP_NUM_THREADS']))

    args.distributed = args.num_gpus > 1
    args.device = torch.device(args.device)
    
    mkdir(args.output_dir)
    logger = setup_logger("Graphormer", args.output_dir, get_rank())
    set_seed(args.seed, args.num_gpus)
    logger.info("Using {} GPUs".format(args.num_gpus))

    # Mesh and SMPL utils
    smpl = SMPL().to(args.device)
    mesh_sampler = Mesh()

    # Renderer for visualization
    renderer = Renderer(faces=smpl.faces.cpu().numpy())

    # Load model
    trans_encoder = []

    input_feat_dim = [int(item) for item in args.input_feat_dim.split(',')]
    hidden_feat_dim = [int(item) for item in args.hidden_feat_dim.split(',')]
    output_feat_dim = input_feat_dim[1:] + [3]

    # which encoder block to have graph convs
    which_blk_graph = [int(item) for item in args.which_gcn.split(',')]
    
    if args.run_eval_only==True and args.resume_checkpoint!=None and args.resume_checkpoint!='None' and 'state_dict' not in args.resume_checkpoint:
        # if only run eval, load checkpoint
        logger.info("Evaluation: Loading from checkpoint {}".format(args.resume_checkpoint))
        _model = torch.load(args.resume_checkpoint)
    else:
        # init three transformer-encoder blocks in a loop
        for i in range(len(output_feat_dim)):
            config_class, model_class = BertConfig, Graphormer
            config = config_class.from_pretrained(args.config_name if args.config_name \
                    else args.model_name_or_path)

            config.output_attentions = False
            config.img_feature_dim = input_feat_dim[i] 
            config.output_feature_dim = output_feat_dim[i]
            args.hidden_size = hidden_feat_dim[i]
            args.intermediate_size = int(args.hidden_size*args.interm_size_scale)

            if which_blk_graph[i]==1:
                config.graph_conv = True
                logger.info("Add Graph Conv")
            else:
                config.graph_conv = False

            config.mesh_type = args.mesh_type

            # update model structure if specified in arguments
            update_params = ['num_hidden_layers', 'hidden_size', 'num_attention_heads', 'intermediate_size']

            for idx, param in enumerate(update_params):
                arg_param = getattr(args, param)
                config_param = getattr(config, param)
                if arg_param > 0 and arg_param != config_param:
                    logger.info("Update config parameter {}: {} -> {}".format(param, config_param, arg_param))
                    setattr(config, param, arg_param)

            # init a transformer encoder and append it to a list
            assert config.hidden_size % config.num_attention_heads == 0
            model = model_class(config=config) 
            logger.info("Init model from scratch.")
            trans_encoder.append(model)

        # init ImageNet pre-trained backbone model
        if args.arch=='hrnet':
            hrnet_yaml = 'models/hrnet/cls_hrnet_w40_sgd_lr5e-2_wd1e-4_bs32_x100.yaml'
            hrnet_checkpoint = 'models/hrnet/hrnetv2_w40_imagenet_pretrained.pth'
            hrnet_update_config(hrnet_config, hrnet_yaml)
            backbone = get_cls_net_gridfeat(hrnet_config, pretrained=hrnet_checkpoint)
            logger.info('=> loading hrnet-v2-w40 model')
        elif args.arch=='hrnet-w64':
            hrnet_yaml = 'models/hrnet/cls_hrnet_w64_sgd_lr5e-2_wd1e-4_bs32_x100.yaml'
            hrnet_checkpoint = 'models/hrnet/hrnetv2_w64_imagenet_pretrained.pth'
            hrnet_update_config(hrnet_config, hrnet_yaml)
            backbone = get_cls_net_gridfeat(hrnet_config, pretrained=hrnet_checkpoint)
            logger.info('=> loading hrnet-v2-w64 model')
        else:
            print("=> using pre-trained model '{}'".format(args.arch))
            backbone = models.__dict__[args.arch](pretrained=True)
            # remove the last fc layer
            backbone = torch.nn.Sequential(*list(backbone.children())[:-2])


        trans_encoder = torch.nn.Sequential(*trans_encoder)
        total_params = sum(p.numel() for p in trans_encoder.parameters())
        logger.info('Graphormer encoders total parameters: {}'.format(total_params))
        backbone_total_params = sum(p.numel() for p in backbone.parameters())
        logger.info('Backbone total parameters: {}'.format(backbone_total_params))

        # build end-to-end Graphormer network (CNN backbone + multi-layer graphormer encoder)
        _model = Graphormer_Network(args, config, backbone, trans_encoder, mesh_sampler)

        if args.resume_checkpoint!=None and args.resume_checkpoint!='None':
            # for fine-tuning or resume training or inference, load weights from checkpoint
            logger.info("Loading state dict from checkpoint {}".format(args.resume_checkpoint))
            # workaround approach to load sparse tensor in graph conv.
            states = torch.load(args.resume_checkpoint)
            # states = checkpoint_loaded.state_dict()
            for k, v in states.items():
                states[k] = v.cpu()
            # del checkpoint_loaded
            _model.load_state_dict(states, strict=False)
            del states
            gc.collect()
            torch.cuda.empty_cache()

    # update configs to enable attention outputs
    setattr(_model.trans_encoder[-1].config,'output_attentions', True)
    setattr(_model.trans_encoder[-1].config,'output_hidden_states', True)
    _model.trans_encoder[-1].bert.encoder.output_attentions = True
    _model.trans_encoder[-1].bert.encoder.output_hidden_states =  True
    for iter_layer in range(4):
        _model.trans_encoder[-1].bert.encoder.layer[iter_layer].attention.self.output_attentions = True
    for inter_block in range(3):
        setattr(_model.trans_encoder[-1].config,'device', args.device)

    _model.to(args.device)
    logger.info("Run inference")

    image_list = []
    if not args.image_file_or_path:
        raise ValueError("image_file_or_path not specified")
    if op.isfile(args.image_file_or_path):
        image_list = [args.image_file_or_path]
    elif op.isdir(args.image_file_or_path):
        # should be a path with images only
        for filename in os.listdir(args.image_file_or_path):
            if filename.endswith(".png") or filename.endswith(".jpg") and 'pred' not in filename:
                image_list.append(args.image_file_or_path+'/'+filename) 
    else:
        raise ValueError("Cannot find images at {}".format(args.image_file_or_path))

    run_inference(args, image_list, _model, smpl, renderer, mesh_sampler)   

if __name__ == "__main__":
    args = parse_args()
    main(args)



from ..utils import common_annotator_call, create_node_input_types
import comfy.model_management as model_management

class PIDINET_Preprocessor:
    @classmethod
    def INPUT_TYPES(s):
        return create_node_input_types(
            safe=(["enable", "disable"], {"default": "enable"})
        )

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "execute"

    CATEGORY = "ControlNet Preprocessors/Line Extractors"

    def execute(self, image, safe, resolution=512, **kwargs):
        from controlnet_aux.pidi import PidiNetDetector

        model = PidiNetDetector.from_pretrained().to(model_management.get_torch_device())
        out = common_annotator_call(model, image, resolution=resolution, safe = safe == "enable")
        del model
        return (out, )

NODE_CLASS_MAPPINGS = {
    "PiDiNetPreprocessor": PIDINET_Preprocessor,
}
NODE_DISPLAY_NAME_MAPPINGS = {
    "PiDiNetPreprocessor": "PiDiNet Lines"
}